"""The project's command line launcher."""
import logging

import click

from bluepysnap import circuit_validation


@click.group()
@click.version_option()
@click.option("-v", "--verbose", count=True)
def cli(verbose):
    """The CLI object."""
    logging.basicConfig(
        level=(logging.WARNING, logging.INFO, logging.DEBUG)[min(verbose, 2)],
        format="%(asctime)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )


@cli.command()
@click.argument("config_file", type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option(
    "--skip-slow/--no-skip-slow",
    default=True,
    help=(
        "Skip slow checks; checking all edges refer to existing node ids, "
        "edge indices are correct, etc"
    ),
)
@click.option("--only-errors", is_flag=True, help="Only print fatal errors (ignore warnings)")
def validate(config_file, skip_slow, only_errors):
    """Validate of Sonata circuit based on config file.

    Args:
        config_file (str): path to Sonata circuit config file
        skip_slow (bool): skip slow tests
        only_errors (bool): only print fatal errors
    """
    circuit_validation.validate(config_file, skip_slow, only_errors)



"""Module providing utility functions for the tests"""

import json
import pickle
import shutil
import tempfile
from contextlib import contextmanager
from distutils.dir_util import copy_tree
from pathlib import Path

import libsonata
import numpy.testing as npt
import pytest

from bluepysnap.circuit import Circuit
from bluepysnap.nodes import NodePopulation, Nodes

TEST_DIR = Path(__file__).resolve().parent
TEST_DATA_DIR = TEST_DIR / "data"

# Pickle size tests often fail when run locally. At the moment, the only thing affecting the
# pickled size is the path length. This is to estimate a safe offset for the size limit
# based on the pickled size of the path of the test data directory.
PICKLED_SIZE_ADJUSTMENT = len(pickle.dumps(str(TEST_DATA_DIR.absolute())))


@contextmanager
def setup_tempdir(cleanup=True):
    temp_dir = str(Path(tempfile.mkdtemp()).resolve())
    try:
        yield temp_dir
    finally:
        if cleanup:
            shutil.rmtree(temp_dir)


@contextmanager
def copy_test_data(config="circuit_config.json"):
    """Copies test/data circuit to a temp directory.

    We don't all data every time but considering this is a copy into a temp dir, it should be fine.
    Returns:
        yields a path to the copy of the config file
    """
    with setup_tempdir() as tmp_dir:
        copy_tree(str(TEST_DATA_DIR), tmp_dir)
        copied_path = Path(tmp_dir)
        yield copied_path, copied_path / config


@contextmanager
def copy_config(config="circuit_config.json"):
    """Copies config to a temp directory.

    Returns:
        yields a path to the copy of the config file
    """
    with setup_tempdir() as tmp_dir:
        output = Path(tmp_dir, config)
        shutil.copy(str(TEST_DATA_DIR / config), str(output))
        yield output


@contextmanager
def edit_config(config_path):
    """Context manager within which you can edit a circuit config. Edits are saved on the context
    manager leave.

    Args:
        config_path (Path): path to config

    Returns:
        Yields a json dict instance of the config_path. This instance will be saved as the config.
    """
    with config_path.open("r") as f:
        config = json.load(f)
    try:
        yield config
    finally:
        with config_path.open("w") as f:
            f.write(json.dumps(config))


def create_node_population(filepath, pop_name, circuit=None, node_sets=None, pop_type=None):
    """Creates a node population.
    Args:
        filepath (str): path to the node file.
        pop_name (str): population name inside the file.
        circuit (Mock/Circuit): either a real circuit or a Mock containing the nodes.
        node_sets: (Mock/NodeSets): either a real node_sets or a mocked node_sets.
        pop_type (str): optional population type.
    Returns:
        NodePopulation: return a node population.
    """
    node_pop_config = {
        "nodes_file": filepath,
        "node_types_file": None,
        "populations": {pop_name: {}},
    }

    if pop_type is not None:
        node_pop_config["populations"][pop_name]["type"] = pop_type

    with copy_test_data() as (_, config_path):
        with edit_config(config_path) as config:
            config["networks"]["nodes"] = [node_pop_config]

        if circuit is None:
            circuit = Circuit(config_path)

    if node_sets is not None:
        circuit.node_sets = node_sets

    node_pop = NodePopulation(circuit, pop_name)
    circuit.nodes = Nodes(circuit)
    return node_pop


def assert_array_equal_strict(x, y):
    # With numpy >= 1.22.4 it would be possible to specify strict=True.
    # The strict parameter ensures that the array data types match.
    npt.assert_array_equal(x, y)
    assert x.dtype == y.dtype



import re
from collections.abc import Iterable
from pathlib import Path

import h5py
import numpy as np
import pytest

import bluepysnap.schemas.schemas as test_module
from bluepysnap.circuit_validation import Error

from utils import TEST_DATA_DIR, copy_test_data, edit_config


@pytest.mark.parametrize(
    "type_args",
    (
        ["circuit"],
        ["definitions", "datatypes"],
        ["definitions", "node"],
        ["definitions", "edge"],
        ["node", "biophysical"],
        ["node", "virtual"],
        ["edge", "chemical"],
        ["edge", "electrical"],
        ["edge", "chemical_virtual"],
    ),
)
def test__load_schema_file_exists(type_args):
    assert isinstance(test_module._load_schema_file(*type_args), dict)


def test__load_schema_file_not_existent():
    fake_type = "fake_type"
    with pytest.raises(FileNotFoundError, match=f"Schema file .*{fake_type}.yaml not found"):
        test_module._load_schema_file(fake_type)


def test__validate_schema_for_dict_correct():
    schema = {"properties": {"test": {"required": ["required_field"]}}}
    to_test = {"test": {"required_field": 42}}

    res = test_module._validate_schema_for_dict(schema, to_test)
    assert isinstance(res, Iterable)
    assert len([*res]) == 0


def test__validate_schema_for_dict_incorrect():
    schema = {"properties": {"test": {"required": ["required_field"]}}}
    to_test = {"test": {"these_aren't_the_droids_you're_looking_for": 666}}

    res = test_module._validate_schema_for_dict(schema, to_test)
    assert isinstance(res, Iterable)
    res = [*res]
    assert len(res) == 1
    assert res[0].message == "'required_field' is a required property"


def test__parse_schema_circuit():
    res = test_module._parse_schema("circuit")
    assert "$typedefs" not in res
    assert "$node_file_defs" not in res
    assert "Circuit" in res["title"]


def test__parse_schema_nodes():
    sub_type = "virtual"
    res = test_module._parse_schema("node", sub_type)
    assert "$typedefs" in res
    assert "$node_file_defs" in res
    assert sub_type in res["title"]


def test__parse_schema_edges():
    sub_type = "electrical"
    res = test_module._parse_schema("edge", sub_type)
    assert "$typedefs" in res
    assert "$edge_file_defs" in res
    assert sub_type in res["title"]


def test__parse_schema_unknown():
    obj_type = "fake"
    with pytest.raises(RuntimeError, match=f"Unknown object type: {obj_type}"):
        test_module._parse_schema(obj_type, "")


def test__get_h5_structure_as_dict(tmp_path):
    expected = {
        "group_1": {
            "dataset_1": {"datatype": "int32"},
            "group_2": {
                "dataset_2": {
                    "datatype": "utf-8",
                    "attributes": {"test_attr": "test_value"},
                }
            },
        }
    }
    with h5py.File(tmp_path / "test.h5", "w") as h5:
        h5.create_group("group_1")
        h5["group_1"].create_dataset("dataset_1", data=[0], dtype="i4")
        h5["group_1"].create_group("group_2")
        h5["group_1/group_2"].create_dataset(
            "dataset_2", data="text", dtype=h5py.string_dtype("utf-8")
        )
        h5["group_1/group_2/dataset_2"].attrs.create("test_attr", "test_value")
        res = test_module._get_h5_structure_as_dict(h5)

    assert res == expected


def test__get_h5_structure_as_dict_library_entries(tmp_path):
    # Check that enumerated values types are resolved from @library
    expected = {
        "@library": {
            "dataset_1": {"datatype": "utf-8"},
            "dataset_2": {"datatype": "utf-8"},
        },
        "dataset_1": {"datatype": "utf-8"},
        "dataset_2": {"datatype": "utf-8"},
        "dataset_3": {"datatype": "int64"},
    }
    with h5py.File(tmp_path / "test.h5", "w") as h5:
        lib = h5.create_group("@library")
        h5.create_dataset("dataset_1", data=[0], dtype="i4")
        h5.create_dataset("dataset_2", data=[0], dtype="u4")
        h5.create_dataset("dataset_3", data=[0], dtype="i8")

        lib.create_dataset("dataset_1", data="some text", dtype=h5py.string_dtype("utf-8"))
        lib.create_dataset("dataset_2", data="also, text", dtype=h5py.string_dtype("utf-8"))

        res = test_module._get_h5_structure_as_dict(h5)

    assert res == expected


def test_validate_config_ok():
    config = str(TEST_DATA_DIR / "circuit_config.json")
    res = test_module.validate_circuit_schema("fake_path", config)

    assert len(res) == 0


@pytest.mark.parametrize(
    "to_remove",
    (
        ["components"],
        ["networks", "nodes", 0, "populations", "default", "type"],
        ["networks", "edges", 0, "populations", "default", "type"],
    ),
)
def test_validate_config_ok_missing_optional_fields(to_remove):
    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            c = config
            for key in to_remove[:-1]:
                c = c[key]
            del c[to_remove[-1]]
        errors = test_module.validate_circuit_schema(str(config_copy_path), config)

        assert len(errors) == 0


@pytest.mark.parametrize(
    ("to_remove_list", "expected"),
    (
        (
            [
                ["networks", "nodes", 0, "populations", "default"],
                ["networks", "nodes", 0, "populations", "default2"],
            ],
            "networks.nodes[0].populations: too few properties",
        ),
        (
            [
                ["networks", "edges", 0, "populations", "default"],
                ["networks", "edges", 0, "populations", "default2"],
            ],
            "networks.edges[0].populations: too few properties",
        ),
        (
            [["networks", "nodes", 0, "populations"]],
            "networks.nodes[0]: 'populations' is a required property",
        ),
        (
            [["networks", "edges", 0, "populations"]],
            "networks.edges[0]: 'populations' is a required property",
        ),
        ([["networks", "nodes", 0]], "networks.nodes: [] is too short"),
        ([["networks", "edges", 0]], "networks.edges: [] is too short"),
        ([["networks", "nodes"]], "networks: 'nodes' is a required property"),
        ([["networks", "edges"]], "networks: 'edges' is a required property"),
        ([["networks"]], "'networks' is a required property"),
    ),
)
def test_validate_config_error(to_remove_list, expected):
    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            for to_remove in to_remove_list:
                c = config
                for key in to_remove[:-1]:
                    c = c[key]
                del c[to_remove[-1]]
        errors = test_module.validate_circuit_schema(str(config_copy_path), config)

        assert len(errors) == 1
        assert errors[0] == Error(Error.FATAL, f"{config_copy_path}:\n\t{expected}")


def test_validate_nodes_ok():
    errors = test_module.validate_nodes_schema(
        str(TEST_DATA_DIR / "nodes_single_pop.h5"), "biophysical"
    )
    assert len(errors) == 0


def test_validate_edges_ok():
    errors = test_module.validate_edges_schema(
        str(TEST_DATA_DIR / "edges_single_pop.h5"), "chemical", virtual=False
    )
    assert len(errors) == 0


@pytest.mark.parametrize(
    "missing",
    (
        "nodes",
        "nodes/default/0",
        "nodes/default/0/dynamics_params",
        "nodes/default/0/dynamics_params/holding_current",
        "nodes/default/0/dynamics_params/threshold_current",
        "nodes/default/0/etype",
        "nodes/default/0/model_template",
        "nodes/default/0/model_type",
        "nodes/default/0/morph_class",
        "nodes/default/0/morphology",
        "nodes/default/0/mtype",
        "nodes/default/0/orientation_w",
        "nodes/default/0/orientation_x",
        "nodes/default/0/orientation_y",
        "nodes/default/0/orientation_z",
        "nodes/default/0/synapse_class",
        "nodes/default/0/x",
        "nodes/default/0/y",
        "nodes/default/0/z",
        "nodes/default/node_type_id",
    ),
)
def test_validate_nodes_biophysical_missing_required(missing):
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        nodes_file = circuit_copy_path / "nodes_single_pop.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            del h5f[missing]
        errors = test_module.validate_nodes_schema(str(nodes_file), "biophysical")
        assert len(errors) == 1
        assert f"'{missing.split('/')[-1]}' is a required property" in errors[0].message


@pytest.mark.parametrize(
    "missing",
    (
        "edges",
        "edges/default/0",
        "edges/default/0/afferent_center_x",
        "edges/default/0/afferent_center_y",
        "edges/default/0/afferent_center_z",
        "edges/default/0/afferent_section_id",
        "edges/default/0/afferent_section_pos",
        "edges/default/0/afferent_section_type",
        "edges/default/0/afferent_segment_id",
        "edges/default/0/afferent_segment_offset",
        "edges/default/0/afferent_surface_x",
        "edges/default/0/afferent_surface_y",
        "edges/default/0/afferent_surface_z",
        "edges/default/0/conductance",
        "edges/default/0/decay_time",
        "edges/default/0/delay",
        "edges/default/0/depression_time",
        "edges/default/0/efferent_center_x",
        "edges/default/0/efferent_center_y",
        "edges/default/0/efferent_center_z",
        "edges/default/0/efferent_section_id",
        "edges/default/0/efferent_section_pos",
        "edges/default/0/efferent_section_type",
        "edges/default/0/efferent_segment_id",
        "edges/default/0/efferent_segment_offset",
        "edges/default/0/efferent_surface_x",
        "edges/default/0/efferent_surface_y",
        "edges/default/0/efferent_surface_z",
        "edges/default/0/facilitation_time",
        "edges/default/0/n_rrp_vesicles",
        "edges/default/0/spine_length",
        "edges/default/0/syn_type_id",
        "edges/default/0/u_syn",
        "edges/default/edge_type_id",
        "edges/default/source_node_id",
        "edges/default/target_node_id",
    ),
)
def test_validate_edges_chemical_missing_required(missing):
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        edges_file = circuit_copy_path / "edges_single_pop.h5"
        with h5py.File(edges_file, "r+") as h5f:
            del h5f[missing]
        errors = test_module.validate_edges_schema(str(edges_file), "chemical", virtual=False)
        assert len(errors) == 1
        assert f"'{missing.split('/')[-1]}' is a required property" in errors[0].message


def test_missing_edge_population():
    with copy_test_data() as (circuit_copy_path, _):
        edges_file = circuit_copy_path / "edges_single_pop.h5"
        with h5py.File(edges_file, "r+") as h5f:
            del h5f["edges/default"]
        errors = test_module.validate_edges_schema(str(edges_file), "chemical", virtual=False)
        assert len(errors) == 1
        assert errors[0] == Error(Error.FATAL, f"{str(edges_file)}:\n\tedges: too few properties")


def test_missing_node_population():
    with copy_test_data() as (circuit_copy_path, _):
        nodes_file = circuit_copy_path / "nodes_single_pop.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            del h5f["nodes/default"]
        errors = test_module.validate_nodes_schema(str(nodes_file), "biophysical")
        assert len(errors) == 1
        assert errors[0] == Error(Error.FATAL, f"{str(nodes_file)}:\n\tnodes: too few properties")


def test_2_edge_populations():
    with copy_test_data() as (circuit_copy_path, _):
        edges_file = circuit_copy_path / "edges_single_pop.h5"
        with h5py.File(edges_file, "r+") as h5f:
            h5f["edges/default2"] = h5f["edges/default"]
        errors = test_module.validate_edges_schema(str(edges_file), "chemical", virtual=False)
        assert len(errors) == 0


def test_2_node_populations():
    with copy_test_data() as (circuit_copy_path, _):
        nodes_file = circuit_copy_path / "nodes_single_pop.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            h5f["nodes/default2"] = h5f["nodes/default"]
        errors = test_module.validate_nodes_schema(str(nodes_file), "biophysical")
        assert len(errors) == 0


def test_virtual_edge_population_ok():
    with copy_test_data() as (circuit_copy_path, _):
        edges_file = circuit_copy_path / "edges_single_pop.h5"
        to_remove = ["efferent_center_x", "efferent_center_y", "efferent_center_z"]
        with h5py.File(edges_file, "r+") as h5f:
            for r in to_remove:
                del h5f[f"edges/default/0/{r}"]
        errors = test_module.validate_edges_schema(str(edges_file), "chemical", virtual=True)
        assert len(errors) == 0


def test_virtual_edge_population_error():
    with copy_test_data() as (circuit_copy_path, _):
        edges_file = circuit_copy_path / "edges_single_pop.h5"
        with h5py.File(edges_file, "r+") as h5f:
            del h5f[f"edges/default/0/afferent_center_x"]
        errors = test_module.validate_edges_schema(str(edges_file), "chemical", virtual=True)
        assert len(errors) == 1
        assert "'afferent_center_x' is a required property" in errors[0].message


def test_virtual_node_population_ok():
    with copy_test_data() as (circuit_copy_path, _):
        nodes_file = circuit_copy_path / "nodes_single_pop.h5"
        to_remove = ["orientation_w", "orientation_x", "orientation_y", "orientation_z"]
        with h5py.File(nodes_file, "r+") as h5f:
            for r in to_remove:
                del h5f[f"nodes/default/0/{r}"]
        errors = test_module.validate_nodes_schema(str(nodes_file), "virtual")
        assert len(errors) == 0


def test_virtual_node_population_error():
    with copy_test_data() as (circuit_copy_path, _):
        nodes_file = circuit_copy_path / "nodes_single_pop.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            del h5f["nodes/default/0/model_type"]
        errors = test_module.validate_nodes_schema(str(nodes_file), "virtual")
        assert len(errors) == 1
        assert "'model_type' is a required property" in errors[0].message


def test_validate_edges_missing_attributes_field():
    # has no attributes at all
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        edges_file = circuit_copy_path / "edges_single_pop.h5"
        with h5py.File(edges_file, "r+") as h5f:
            del h5f["edges/default/source_node_id"].attrs["node_population"]
        errors = test_module.validate_edges_schema(str(edges_file), "chemical", virtual=False)
        assert len(errors) == 1
        assert "missing required attribute(s) ['node_population']" in errors[0].message


def test_validate_edges_missing_attribute():
    # has attributes but not the required ones
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        edges_file = circuit_copy_path / "edges_single_pop.h5"
        with h5py.File(edges_file, "r+") as h5f:
            del h5f["edges/default/source_node_id"].attrs["node_population"]
            h5f["edges/default/source_node_id"].attrs.create("population", "some_val")
        errors = test_module.validate_edges_schema(str(edges_file), "chemical", virtual=False)
        assert len(errors) == 1
        assert "missing required attribute(s) ['node_population']" in errors[0].message


@pytest.mark.parametrize(
    "field",
    (
        "edges/default/0/afferent_center_x",
        "edges/default/0/afferent_center_y",
        "edges/default/0/afferent_center_z",
        "edges/default/edge_type_id",
    ),
)
def test_wrong_datatype(field):
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        edges_file = circuit_copy_path / "edges_single_pop.h5"
        with h5py.File(edges_file, "r+") as h5f:
            del h5f[field]
            h5f.create_dataset(field, data=[0], dtype="i2")
        errors = test_module.validate_edges_schema(str(edges_file), "chemical", virtual=False)
        assert len(errors) == 1
        assert errors[0].level == Error.WARNING
        assert f"incorrect datatype 'int16' for '{field}'" in errors[0].message


def test__get_reference_resolver():
    expected = {"const": "test_value"}
    schema = {"$mock_reference": expected}
    resolver = test_module._get_reference_resolver(schema)
    assert resolver.lookup("#/$mock_reference").contents == expected


def test_nodes_schema_types():
    property_types, dynamics_params = test_module.nodes_schema_types("biophysical")
    assert "x" in property_types
    assert property_types["x"] == np.float32


def test_edges_schema_types():
    edge_property_types = test_module.edges_schema_types("chemical", virtual=True)
    assert "afferent_center_x" in edge_property_types
    assert edge_property_types["afferent_center_x"] == np.float32



from unittest.mock import patch

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest

import bluepysnap.frame_report as test_module
from bluepysnap.bbp import Cell
from bluepysnap.circuit_ids import CircuitNodeId, CircuitNodeIds
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.simulation import Simulation
from bluepysnap.utils import IDS_DTYPE

from utils import TEST_DATA_DIR


class TestFrameReport:
    def setup_method(self):
        self.simulation = Simulation(str(TEST_DATA_DIR / "simulation_config.json"))
        self.test_obj = test_module.FrameReport(self.simulation, "soma_report")
        self.test_obj_info = test_module.FrameReport(self.simulation, "section_report")

    def test_config(self):
        assert self.test_obj.config == {
            "cells": "Layer23",
            "variable_name": "m",
            "sections": "soma",
            "type": "compartment",
            "file_name": "soma_report",
            "start_time": 0,
            "end_time": 1000.0,
            "dt": 0.01,
            "enabled": True,
        }

    def test_time_start(self):
        assert self.test_obj.time_start == 0.0
        assert self.test_obj_info.time_start == 0.2

    def test_time_stop(self):
        assert self.test_obj.time_stop == 1000.0
        assert self.test_obj_info.time_stop == 0.8

    def test_dt(self):
        assert self.test_obj.dt == 0.01
        with patch("bluepysnap.frame_report.L") as log_mock:
            assert self.test_obj_info.dt == 0.02
            assert log_mock.warning.call_count == 1

    def test_time_units(self):
        assert self.test_obj.time_units == "ms"
        with pytest.raises(BluepySnapError):
            self.test_obj_info.time_units

    def test_data_units(self):
        assert self.test_obj.data_units == "mV"
        with pytest.raises(BluepySnapError):
            self.test_obj_info.data_units

    def test_sim(self):
        assert isinstance(self.test_obj.simulation, Simulation)
        assert isinstance(self.test_obj_info.simulation, Simulation)

    def test_node_set(self):
        assert self.test_obj.node_set == "Layer23"
        assert self.test_obj_info.node_set == "Layer23"

    def test_population_names(self):
        assert sorted(self.test_obj.population_names) == ["default", "default2"]
        assert sorted(self.test_obj_info.population_names) == ["default", "default2"]

    def test_get_population(self):
        assert isinstance(self.test_obj["default"], test_module.PopulationFrameReport)
        assert isinstance(self.test_obj_info["default"], test_module.PopulationFrameReport)

    def test_iter(self):
        assert sorted(list(self.test_obj)) == ["default", "default2"]
        for report in self.test_obj:
            isinstance(report, test_module.PopulationFrameReport)

        assert sorted(list(self.test_obj_info)) == ["default", "default2"]
        for report in self.test_obj_info:
            isinstance(report, test_module.PopulationFrameReport)


class TestCompartmentReport:
    def setup_method(self):
        self.simulation = Simulation(str(TEST_DATA_DIR / "simulation_config.json"))
        self.test_obj = test_module.CompartmentReport(self.simulation, "section_report")

    def test_get_population(self):
        assert isinstance(self.test_obj["default"], test_module.PopulationCompartmentReport)

    def test_iter(self):
        assert sorted(list(self.test_obj)) == ["default", "default2"]
        for report in self.test_obj:
            isinstance(report, test_module.PopulationCompartmentReport)

    def test_filter(self):
        expected = pd.DataFrame(
            data=[
                np.array([0.3, 1.3, 2.3, 3.3, 4.3, 5.3, 6.3] * 2, dtype=np.float32) + 0.1 * i
                for i in range(4)
            ],
            columns=pd.MultiIndex.from_tuples(
                [
                    ("default", 0, 0),
                    ("default", 0, 1),
                    ("default", 1, 0),
                    ("default", 1, 1),
                    ("default", 2, 0),
                    ("default", 2, 1),
                    ("default", 2, 1),
                    ("default2", 0, 0),
                    ("default2", 0, 1),
                    ("default2", 1, 0),
                    ("default2", 1, 1),
                    ("default2", 2, 0),
                    ("default2", 2, 1),
                    ("default2", 2, 1),
                ]
            ),
            index=np.array([0.3, 0.4, 0.5, 0.6]),
        )

        filtered = self.test_obj.filter(group=[0], t_start=0.3, t_stop=0.6)
        assert filtered.frame_report == self.test_obj
        assert filtered.t_start == 0.3
        assert filtered.t_stop == 0.6
        assert filtered.group == [0]
        assert isinstance(filtered, test_module.FilteredFrameReport)
        expected_columns = [
            ("default", 0, 0),
            ("default", 0, 1),
            ("default2", 0, 0),
            ("default2", 0, 1),
        ]
        pdt.assert_frame_equal(filtered.report, expected.loc[:, expected_columns])

        filtered = self.test_obj.filter(group={"other1": ["B"]}, t_start=0.3, t_stop=0.6)
        expected_columns = [("default2", 1, 0), ("default2", 1, 1)]
        pdt.assert_frame_equal(filtered.report, expected.loc[:, expected_columns])

        filtered = self.test_obj.filter(group={"population": "default2"}, t_start=0.3, t_stop=0.6)
        expected_columns = [
            ("default2", 0, 0),
            ("default2", 0, 1),
            ("default2", 1, 0),
            ("default2", 1, 1),
            ("default2", 2, 0),
            ("default2", 2, 1),
        ]
        pdt.assert_frame_equal(filtered.report, expected.loc[:, expected_columns])

        filtered = self.test_obj.filter(group={"population": "default3"}, t_start=0.3, t_stop=0.6)
        pdt.assert_frame_equal(filtered.report, expected.iloc[:0, :0])


class TestSomaReport:
    def setup_method(self):
        self.simulation = Simulation(str(TEST_DATA_DIR / "simulation_config.json"))
        self.test_obj = test_module.SomaReport(self.simulation, "soma_report")

    def test_get_population(self):
        assert isinstance(self.test_obj["default"], test_module.PopulationSomaReport)

    def test_iter(self):
        assert sorted(list(self.test_obj)) == ["default", "default2"]
        for report in self.test_obj:
            isinstance(report, test_module.PopulationSomaReport)

    def test_filter(self):
        expected = pd.DataFrame(
            data=[
                np.array([0.3, 1.3, 2.3, 0.3, 1.3, 2.3], dtype=np.float32) + 0.1 * i
                for i in range(4)
            ],
            columns=pd.MultiIndex.from_tuples(
                [
                    ("default", 0),
                    ("default", 1),
                    ("default", 2),
                    ("default2", 0),
                    ("default2", 1),
                    ("default2", 2),
                ]
            ),
            index=np.array([0.3, 0.4, 0.5, 0.6]),
        )

        filtered = self.test_obj.filter(group=None, t_start=0.3, t_stop=0.6)
        assert filtered.frame_report == self.test_obj
        assert filtered.t_start == 0.3
        assert filtered.t_stop == 0.6
        assert filtered.group is None
        assert isinstance(filtered, test_module.FilteredFrameReport)
        pdt.assert_frame_equal(filtered.report, expected)

        filtered = self.test_obj.filter(group={"other1": ["B"]}, t_start=0.3, t_stop=0.6)
        pdt.assert_frame_equal(filtered.report, expected.loc[:, [("default2", 1)]])

        filtered = self.test_obj.filter(group={"population": "default2"}, t_start=0.3, t_stop=0.6)
        pdt.assert_frame_equal(filtered.report, expected.loc[:, ["default2"]])

        filtered = self.test_obj.filter(group={"population": "default3"}, t_start=0.3, t_stop=0.6)
        pdt.assert_frame_equal(filtered.report, expected.iloc[:0, :0])

        ids = CircuitNodeIds.from_arrays(["default", "default", "default2"], [0, 1, 1])
        filtered = self.test_obj.filter(group=ids, t_start=0.3, t_stop=0.6)
        expected_columns = [("default", 0), ("default", 1), ("default2", 1)]
        pdt.assert_frame_equal(filtered.report, expected.loc[:, expected_columns])


class TestPopulationFrameReport:
    def setup_method(self):
        self.simulation = Simulation(str(TEST_DATA_DIR / "simulation_config.json"))
        self.test_obj = test_module.FrameReport(self.simulation, "section_report")["default"]

    def test_name(self):
        assert self.test_obj.name == "default"

    def test__resolve(self):
        with pytest.raises(NotImplementedError):
            self.test_obj._resolve([1])


class TestPopulationCompartmentReport:
    def setup_method(self):
        self.simulation = Simulation(str(TEST_DATA_DIR / "simulation_config.json"))
        self.test_obj = test_module.CompartmentReport(self.simulation, "section_report")["default"]
        timestamps = np.linspace(0, 0.9, 10)
        data = np.array([np.arange(7) + j * 0.1 for j in range(10)], dtype=np.float32)
        ids = [(0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1), (2, 1)]
        self.df = pd.DataFrame(data=data, columns=pd.MultiIndex.from_tuples(ids), index=timestamps)

    @property
    def empty_df(self):
        """Return an empty DataFrame with the original types of index and columns."""
        return self.df.iloc[:0, :0]

    def test__resolve(self):
        npt.assert_array_equal(self.test_obj._resolve({Cell.MTYPE: "L6_Y"}), [1, 2])
        assert self.test_obj._resolve({Cell.MTYPE: "L2_X"}) == [0]
        npt.assert_array_equal(self.test_obj._resolve("Node12_L6_Y"), [1, 2])

    def test_nodes(self):
        assert self.test_obj.nodes.get(group=2, properties=Cell.MTYPE) == "L6_Y"

    def test_nodes_invalid_population(self):
        test_obj = self.test_obj.__class__(self.test_obj.frame_report, self.test_obj.name)
        test_obj._population_name = "unknown"
        with pytest.raises(BluepySnapError):
            test_obj.nodes

    @pytest.mark.parametrize("t_step", [None, 0.02, 0.04, 0.0401, 0.0399, 0.05, 200000])
    def test_get(self, t_step):
        def _assert_frame_equal(df1, df2):
            # compare df1 and df2, after filtering df2 according to t_stride
            df2 = df2.iloc[::t_stride]
            pdt.assert_frame_equal(df1, df2)

        # calculate the expected t_stride, depending on t_step and dt (varying across tests)
        t_stride = round(t_step / self.test_obj.frame_report.dt) if t_step is not None else 1

        _assert_frame_equal(self.test_obj.get(t_step=t_step), self.df)
        _assert_frame_equal(self.test_obj.get([], t_step=t_step), self.empty_df)
        _assert_frame_equal(self.test_obj.get(np.array([]), t_step=t_step), self.empty_df)
        _assert_frame_equal(self.test_obj.get((), t_step=t_step), self.empty_df)

        _assert_frame_equal(self.test_obj.get(2, t_step=t_step), self.df.loc[:, [2]])
        _assert_frame_equal(
            self.test_obj.get(CircuitNodeId("default", 2), t_step=t_step), self.df.loc[:, [2]]
        )

        # not from this population
        _assert_frame_equal(
            self.test_obj.get(CircuitNodeId("default2", 2), t_step=t_step), self.empty_df
        )

        _assert_frame_equal(self.test_obj.get([2, 0], t_step=t_step), self.df.loc[:, [0, 2]])

        _assert_frame_equal(self.test_obj.get([0, 2], t_step=t_step), self.df.loc[:, [0, 2]])

        _assert_frame_equal(
            self.test_obj.get(np.asarray([0, 2]), t_step=t_step), self.df.loc[:, [0, 2]]
        )

        _assert_frame_equal(
            self.test_obj.get([2], t_stop=0.5, t_step=t_step), self.df.iloc[:6].loc[:, [2]]
        )

        _assert_frame_equal(
            self.test_obj.get([2], t_stop=0.55, t_step=t_step), self.df.iloc[:6].loc[:, [2]]
        )

        _assert_frame_equal(
            self.test_obj.get([2], t_start=0.5, t_step=t_step), self.df.iloc[5:].loc[:, [2]]
        )

        _assert_frame_equal(
            self.test_obj.get([2], t_start=0.5, t_stop=0.8, t_step=t_step),
            self.df.iloc[5:9].loc[:, [2]],
        )

        _assert_frame_equal(
            self.test_obj.get([2, 1], t_start=0.5, t_stop=0.8, t_step=t_step),
            self.df.iloc[5:9].loc[:, [1, 2]],
        )

        _assert_frame_equal(
            self.test_obj.get([2, 1], t_start=0.2, t_stop=0.8, t_step=t_step),
            self.df.iloc[2:9].loc[:, [1, 2]],
        )

        _assert_frame_equal(
            self.test_obj.get(group={Cell.MTYPE: "L6_Y"}, t_start=0.2, t_stop=0.8, t_step=t_step),
            self.df.iloc[2:9].loc[:, [1, 2]],
        )

        _assert_frame_equal(
            self.test_obj.get(group={Cell.MTYPE: "L2_X"}, t_step=t_step), self.df.loc[:, [0]]
        )

        _assert_frame_equal(self.test_obj.get(group="Layer23", t_step=t_step), self.df.loc[:, [0]])

        ids = CircuitNodeIds.from_arrays(["default", "default", "default2"], [0, 2, 1])
        _assert_frame_equal(self.test_obj.get(group=ids, t_step=t_step), self.df.loc[:, [0, 2]])

        with pytest.raises(
            BluepySnapError, match="All node IDs must be >= 0 and < 3 for population 'default'"
        ):
            self.test_obj.get(-1, t_start=0.2, t_step=t_step)

        with pytest.raises(BluepySnapError, match="Times cannot be negative"):
            self.test_obj.get(0, t_start=-1, t_step=t_step)

        with pytest.raises(BluepySnapError, match="tstart is after the end of the range"):
            self.test_obj.get([0, 2], t_start=15, t_step=t_step)

        with pytest.raises(
            BluepySnapError, match="All node IDs must be >= 0 and < 3 for population 'default'"
        ):
            self.test_obj.get(4, t_step=t_step)

    @pytest.mark.parametrize("t_step", [0, -1, 0.0000001])
    def test_get_with_invalid_t_step(self, t_step):
        match = f"Invalid t_step={t_step}. It should be None or a multiple of"
        with pytest.raises(BluepySnapError, match=match):
            self.test_obj.get(t_step=t_step)

    def test_get_partially_not_in_report(self):
        with patch.object(self.test_obj.__class__, "_resolve", return_value=np.asarray([0, 4])):
            pdt.assert_frame_equal(self.test_obj.get([0, 4]), self.df.loc[:, [0]])

    def test_get_not_in_report(self):
        with patch.object(self.test_obj.__class__, "_resolve", return_value=np.asarray([4])):
            pdt.assert_frame_equal(self.test_obj.get([4]), self.empty_df)

    def test_node_ids(self):
        npt.assert_array_equal(self.test_obj.node_ids, np.array(sorted([0, 1, 2]), dtype=IDS_DTYPE))


class TestPopulationSomaReport(TestPopulationCompartmentReport):
    def setup_method(self):
        self.simulation = Simulation(str(TEST_DATA_DIR / "simulation_config.json"))
        self.test_obj = test_module.SomaReport(self.simulation, "soma_report")["default"]
        timestamps = np.linspace(0, 0.9, 10)
        data = {0: timestamps, 1: timestamps + 1, 2: timestamps + 2}
        self.df = pd.DataFrame(data=data, index=timestamps, columns=[0, 1, 2]).astype(np.float32)



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Access to circuit data."""
import logging
from pathlib import Path

from cached_property import cached_property

from bluepysnap.config import CircuitConfig, CircuitConfigStatus
from bluepysnap.edges import Edges
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.node_sets import NodeSets
from bluepysnap.nodes import Nodes

L = logging.getLogger(__name__)


class Circuit:
    """Access to circuit data."""

    def __init__(self, config):
        """Initializes a circuit object from a SONATA config file.

        Args:
            config (str): Path to a SONATA config file.

        Returns:
            Circuit: A Circuit object.
        """
        self._circuit_config_path = str(Path(config).absolute())
        self._config = CircuitConfig.from_config(config)

        if self.partial_config:
            L.info(
                "Loaded PARTIAL circuit config. Functionality may be limited. "
                "It is up to the user to be diligent when accessing properties."
            )

    @property
    def to_libsonata(self):
        """Libsonata instance of the circuit."""
        return self._config.to_libsonata

    @property
    def config(self):
        """Network config dictionary."""
        return self._config.to_dict()

    def get_node_population_config(self, name):
        """Get node population configuration."""
        try:
            return self._config.node_populations[name]
        except KeyError as e:
            raise BluepySnapError(f"Population config not found for node population: {name}") from e

    def get_edge_population_config(self, name):
        """Get edge population configuration."""
        try:
            return self._config.edge_populations[name]
        except KeyError as e:
            raise BluepySnapError(f"Population config not found for edge population: {name}") from e

    @cached_property
    def node_sets(self):
        """Returns the NodeSets object bound to the circuit."""
        path = self.to_libsonata.node_sets_path
        return NodeSets.from_file(path) if path else NodeSets.from_dict({})

    @cached_property
    def nodes(self):
        """Access to node population(s). See :py:class:`~bluepysnap.nodes.Nodes`."""
        return Nodes(self)

    @cached_property
    def edges(self):
        """Access to edge population(s). See :py:class:`~bluepysnap.edges.Edges`."""
        return Edges(self)

    @cached_property
    def partial_config(self):
        """Check partiality of the config."""
        return self._config.status == CircuitConfigStatus.partial

    def __getstate__(self):
        """Make Circuits pickle-able, without storing state of caches."""
        return self._circuit_config_path

    def __setstate__(self, state):
        """Load from pickle state."""
        self.__init__(state)



import numpy as np
import numpy.testing as npt
import pytest

import bluepysnap.utils as test_module
from bluepysnap.circuit_ids import CircuitEdgeId, CircuitNodeId
from bluepysnap.exceptions import (
    BluepySnapDeprecationError,
    BluepySnapDeprecationWarning,
    BluepySnapError,
)
from bluepysnap.sonata_constants import DYNAMICS_PREFIX

from utils import TEST_DATA_DIR


def test_load_json():
    actual = test_module.load_json(str(TEST_DATA_DIR / "circuit_config.json"))
    assert actual["manifest"]["$BASE_DIR"] == "."


def test_is_iterable():
    assert test_module.is_iterable([12, 13])
    assert test_module.is_iterable(np.asarray([12, 13]))
    assert not test_module.is_iterable(12)
    assert not test_module.is_iterable("abc")


def test_is_node_id():
    assert test_module.is_node_id(1)
    assert test_module.is_node_id(np.int32(1))
    assert test_module.is_node_id(np.uint32(1))
    assert test_module.is_node_id(np.int64(1))
    assert test_module.is_node_id(np.uint64(1))
    assert test_module.is_node_id(CircuitNodeId("default", 1))

    assert not test_module.is_node_id([1])
    assert not test_module.is_node_id(np.array([1], dtype=np.int64))
    assert not test_module.is_node_id(CircuitEdgeId("default", 1))


def test_ensure_list():
    assert test_module.ensure_list(1) == [1]
    assert test_module.ensure_list([1]) == [1]
    assert test_module.ensure_list(iter([1])) == [1]
    assert test_module.ensure_list((2, 1)) == [2, 1]
    assert test_module.ensure_list("abc") == ["abc"]


def test_ensure_ids():
    res = test_module.ensure_ids(np.array([1, 2, 3], dtype=np.uint64))
    npt.assert_equal(res, np.array([1, 2, 3], dtype=test_module.IDS_DTYPE))
    npt.assert_equal(res.dtype, test_module.IDS_DTYPE)


def test_add_dynamic_prefix():
    assert test_module.add_dynamic_prefix(["a", "b"]) == [
        DYNAMICS_PREFIX + "a",
        DYNAMICS_PREFIX + "b",
    ]


def test_euler2mat():
    pi2 = 0.5 * np.pi
    actual = test_module.euler2mat(
        [0.0, pi2],  # rotation_angle_z
        [pi2, 0.0],  # rotation_angle_y
        [pi2, pi2],  # rotation_angle_x
    )
    expected = np.array(
        [
            [
                [0.0, 0.0, 1.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
            ],
            [
                [0.0, -1.0, 0.0],
                [0.0, 0.0, -1.0],
                [1.0, 0.0, 0.0],
            ],
        ]
    )
    npt.assert_almost_equal(actual, expected)

    with pytest.raises(BluepySnapError):
        test_module.euler2mat([pi2, pi2], [pi2, pi2], [pi2])  # ax|y|z not of same size


def test_quaternion2mat():
    actual = test_module.quaternion2mat(
        [1, 1, 1],
        [
            1,
            0,
            0,
        ],
        [0, 1, 0],
        [0, 0, 1],
    )
    expected = np.array(
        [
            [
                [1.0, 0.0, 0.0],
                [0.0, 0.0, -1.0],
                [0.0, 1.0, 0.0],
            ],
            [
                [0.0, 0.0, 1.0],
                [0.0, 1.0, 0.0],
                [-1.0, 0.0, 0.0],
            ],
            [
                [0.0, -1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0],
            ],
        ]
    )
    npt.assert_almost_equal(actual, expected)


class TestDeprecate:
    def test_fail(self):
        with pytest.raises(BluepySnapDeprecationError):
            test_module.Deprecate.fail("something")

    def test_warning(self):
        import warnings

        with warnings.catch_warnings(record=True) as w:
            test_module.Deprecate.warn("something")

            # Verify some things
            assert len(w) == 1
            assert issubclass(w[-1].category, BluepySnapDeprecationWarning)
            assert "something" in str(w[-1].message)



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Module including the sonata node and edge namespaces."""
from bluepysnap.exceptions import BluepySnapError

DYNAMICS_PREFIX = "@dynamics:"
DEFAULT_NODE_TYPE = "biophysical"
DEFAULT_EDGE_TYPE = "chemical"


class ConstContainer:
    """Constant Container for snap.

    Notes:
        Allows the creation of hierarchical subclasses such as:

        .. code-block:: pycon

            >>> class Container(ConstContainer):
            >>>     VAR1 = "var1"
            >>> class SubContainer(Container):
            >>>     VAR2 = "var2"
            >>> class SubSubContainer(SubContainer):
            >>>     VAR3 = "var3"
            >>> print(SubSubContainer.key_set()) # To know the accessible variable names
            {"VAR1", "VAR2", "VAR3"}
            >>> for v in SubSubContainer.key_set(): # To get the variable (names, values)
            >>>     print(v, getattr(SubSubContainer, v))
    """

    @classmethod
    def key_set(cls):
        """List all constant members of the class."""
        all_keys = set()
        for base in cls.__bases__:
            if base is object:
                continue
            try:
                all_keys.update(base.key_set())
            except AttributeError as e:
                raise BluepySnapError(
                    "Container classes must derive from classes implementing key_set method"
                ) from e
        all_keys.update(
            name
            for name in vars(cls)
            if not name.startswith("_") and name not in ["key_set", "get"]
        )
        return all_keys

    @classmethod
    def get(cls, const_name):
        """Get a constant from a string name."""
        try:
            res = getattr(cls, const_name)
        except AttributeError as e:
            raise BluepySnapError("{cls} does not have a '{const_name}' member") from e
        return res


class Node(ConstContainer):
    """Node property names."""

    X = "x"  #:
    Y = "y"  #:
    Z = "z"  #:

    ORIENTATION_W = "orientation_w"  #:
    ORIENTATION_X = "orientation_x"  #:
    ORIENTATION_Y = "orientation_y"  #:
    ORIENTATION_Z = "orientation_z"  #:

    ROTATION_ANGLE_X = "rotation_angle_xaxis"  #:
    ROTATION_ANGLE_Y = "rotation_angle_yaxis"  #:
    ROTATION_ANGLE_Z = "rotation_angle_zaxis"  #:

    MORPHOLOGY = "morphology"  #:

    RECENTER = "recenter"  #:

    MODEL_TYPE = "model_type"  #:
    MODEL_TEMPLATE = "model_template"  #:


class Edge(ConstContainer):
    """Edge property names."""

    SOURCE_NODE_ID = "@source_node"  #:
    TARGET_NODE_ID = "@target_node"  #:

    AXONAL_DELAY = "delay"  #:
    SYN_WEIGHT = "syn_weight"  #:

    POST_SECTION_ID = "afferent_section_id"  #:
    POST_SECTION_POS = "afferent_section_pos"  #:
    PRE_SECTION_ID = "efferent_section_id"  #:
    PRE_SECTION_POS = "efferent_section_pos"  #:

    # postsynaptic touch position (in the center of the segment)
    POST_X_CENTER = "afferent_center_x"  #:
    POST_Y_CENTER = "afferent_center_y"  #:
    POST_Z_CENTER = "afferent_center_z"  #:

    # postsynaptic touch position (on the segment surface)
    POST_X_SURFACE = "afferent_surface_x"  #:
    POST_Y_SURFACE = "afferent_surface_y"  #:
    POST_Z_SURFACE = "afferent_surface_z"  #:

    # presynaptic touch position (in the center of the segment)
    PRE_X_CENTER = "efferent_center_x"  #:
    PRE_Y_CENTER = "efferent_center_y"  #:
    PRE_Z_CENTER = "efferent_center_z"  #:

    # presynaptic touch position (on the segment surface)
    PRE_X_SURFACE = "efferent_surface_x"  #:
    PRE_Y_SURFACE = "efferent_surface_y"  #:
    PRE_Z_SURFACE = "efferent_surface_z"  #:



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Neuron models access."""

from pathlib import Path

from bluepysnap.exceptions import BluepySnapError
from bluepysnap.sonata_constants import Node
from bluepysnap.utils import is_node_id


class NeuronModelsHelper:
    """Collection of neuron models related methods."""

    def __init__(self, properties, population):
        """Constructor.

        Args:
            properties (libsonata.PopulationProperties): properties of the population
            population (NodePopulation): NodePopulation object used to query the nodes.

        Returns:
            NeuronModelsHelper: A NeuronModelsHelper object.
        """
        # all nodes from a population must have the same model type
        if properties.type != "biophysical":
            raise BluepySnapError("Neuron models can be only in biophysical node population.")

        self._properties = properties
        self._population = population

    def get_filepath(self, node_id):
        """Return path to model file corresponding to `node_id`.

        Args:
            node_id (int|CircuitNodeId): node id

        Returns:
            Path: path to the model file of neuron
        """
        if not is_node_id(node_id):
            raise BluepySnapError("node_id must be a int or a CircuitNodeId")
        node = self._population.get(node_id, [Node.MODEL_TYPE, Node.MODEL_TEMPLATE])
        models_dir = self._properties.biophysical_neuron_models_dir

        template = node[Node.MODEL_TEMPLATE]
        assert ":" in template, "Format of 'model_template' must be <schema>:<resource>."
        schema, resource = template.split(":", 1)
        resource = Path(resource).with_suffix(f".{schema}")
        if resource.is_absolute():
            return resource
        return Path(models_dir, resource)



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Exceptions used throughout the library."""


class BluepySnapError(Exception):
    """Base SNAP exception."""


class BluepySnapDeprecationError(Exception):
    """SNAP deprecation exception."""


class BluepySnapDeprecationWarning(DeprecationWarning):
    """SNAP deprecation warning."""



# Copyright (c) 2020, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Circuit ids."""
import abc
from collections import namedtuple

import numpy as np
import pandas as pd

from bluepysnap import utils
from bluepysnap._doctools import AbstractDocSubstitutionMeta
from bluepysnap.exceptions import BluepySnapError

CircuitNodeId = namedtuple("CircuitNodeId", ("population", "id"))
CircuitEdgeId = namedtuple("CircuitEdgeId", ("population", "id"))


class CircuitIds(abc.ABC):
    """High performances CircuitID container.

    This abstract class aims at defining the global ids for the circuit. It is the base class of
    CircuitNodeIds and CircuitEdgeIds.
    The pandas.MultiIndex class is used as backend and can be accessed using CircuitIds.index.
    A global circuit node id is the combination of a population and an ID inside this
    population.
    """

    def __init__(self, index, sort_index=True):
        """Return an instance of CircuitIds.

        Args:
            index (pandas.MultiIndex): a multi index from pandas with the population names as
                first level and node IDs as the second level.
            sort_index (bool): if true sort the index per population and then node IDs.
        """
        if not isinstance(index, pd.MultiIndex):
            raise BluepySnapError("index must be a pandas.MultiIndex object.")
        if sort_index:
            # best perf compared to sort_values. Sorted by population and ids.
            index = index.sortlevel()[0]
        self.index = index

    @classmethod
    def _instance(cls, index, sort_index=True):
        """The instance returned by the functions."""
        return cls(index, sort_index=sort_index)

    @classmethod
    def from_arrays(cls, populations, population_ids, sort_index=True):
        """Create a set of ids using two arrays population(s) and id(s).

        Args:
            populations (list/numpy.array): a sequence of populations. If the population is a
                string then all the ids will be connected to this population.
            population_ids (list/numpy.array): a sequence of node IDs or a single node ID.
            sort_index (bool): will sort the index if set to True. Otherwise the ordering from the
                user inputs is kept. Sorting the index can result in better performances.

        Returns:
            CircuitIds: a set of global node IDs created via the populations and the node IDs
                provided.
        """
        populations = np.asarray(populations)
        population_ids = utils.ensure_ids(population_ids)

        if len(populations) != len(population_ids):
            raise BluepySnapError(
                "populations and population_ids must have the same size. "
                f"{len(populations)} != {len(population_ids)}"
            )

        index = pd.MultiIndex.from_arrays([populations, population_ids])
        return cls(index, sort_index=sort_index)

    @classmethod
    def from_dict(cls, ids_dictionary):
        """Create a set of ids using a dictionary as input.

        Args:
            ids_dictionary (dict): a dictionary with the population as keys and node IDs as
                values.

        Notes:
            with the support of python 2 we cannot guaranty the ordering so we force the sorting
            of ids.

        Returns:
            CircuitIds: a set of global node IDs created via the provided dictionary.

        Examples:
            >>> CircuitIds.from_dict({"pop1": [0,2,4], "pop2": [1,2,5]})
        """
        populations = np.empty((0,), dtype=str)
        population_ids = np.empty((0,), dtype=utils.IDS_DTYPE)
        for population, ids in ids_dictionary.items():
            ids = utils.ensure_ids(ids)
            population_ids = np.append(population_ids, ids)
            populations = np.append(populations, np.full(ids.shape, fill_value=population))
        index = pd.MultiIndex.from_arrays([populations, population_ids])
        return cls(index, sort_index=True)

    @classmethod
    def from_tuples(cls, circuit_id_list, sort_index=True):
        """Create a set of ids using a list of CircuitId or tuples as input.

        Args:
            circuit_id_list (list): a list of CircuitId or list of tuples.
            sort_index (bool): will sort the index if set to True. Otherwise the ordering from the
                user inputs is kept. Sorting the index can result in better performances.

        Returns:
            CircuitIds: a set of global node IDs created via the provided dictionary.

        Examples:
            >>> CircuitIds.from_tuples([("pop1", 0), ("pop1", 2), ("pop2", 0)])
            >>> CircuitIds.from_tuples([CircuitId("pop1", 0), CircuitId("pop1", 2),
            >>>                             CircuitId("pop2", 0)])
        """
        return cls(pd.MultiIndex.from_tuples(circuit_id_list), sort_index=sort_index)

    def _locate(self, population):
        """Returns the index indices corresponding to a given population.

        Args:
            population (str): the population name you want to locate inside the MultiIndex.

        Returns:
            numpy.array: indices corresponding to the population.
        """
        try:
            return self.index.get_locs(utils.ensure_list(population))
        except KeyError:
            return []

    def filter_population(self, population, inplace=False):
        """Filter the IDs corresponding to a population.

        Args:
            population (str): the population you want to extract.
            inplace (bool): if set to True. Do the transformation inplace.

        Returns:
            CircuitIds : a filtered CircuitIds containing only IDs for the given population.
        """
        if not inplace:
            return self._instance(self.index[self._locate(population)], sort_index=False)
        self.index = self.index[self._locate(population)]
        return None

    def get_populations(self, unique=False):
        """Returns all population values from the circuit node IDs."""
        if unique:
            return self.index.levels[0].to_numpy()
        return self.index.get_level_values(0).to_numpy()

    def get_ids(self, unique=False):
        """Returns all the ID values from the circuit node IDs."""
        if unique:
            return self.index.levels[1].to_numpy()
        return self.index.get_level_values(1).to_numpy()

    def copy(self):
        """Copy a CircuitIds."""
        return self._instance(self.index.copy())

    def sort(self, inplace=False):
        """Sort the CircuitIds by population and then by ids.

        Args:
            inplace (bool): if set to True. Do the transformation inplace.

        Notes:
            sorting a CircuitIds will gives better perf for the population extraction and the
            data extraction from a sorted dataframe.
        """
        if not inplace:
            return self._instance(self.index, sort_index=True)
        self.index = self.index.sortlevel()[0]
        return None

    def append(self, other, inplace=False):
        """Append a CircuitIds to the current one.

        Args:
            other (CircuitIds): the other CircuitIds to append to the current one.
            inplace (bool): if set to True. Do the transformation inplace.
        """
        if not inplace:
            return self._instance(self.index.append(other.index))
        self.index = self.index.append(other.index)
        return None

    def _apply(self, fun, inplace):
        """Apply fun on index."""
        res = self if inplace else self.copy()
        res.index = fun(res.index)
        if not inplace:
            return res
        return None

    def sample(self, sample_size, inplace=False):
        """Sample a CircuitIds.

        Notes:
            this function takes randomly ``sample_size`` values of a circuit node ids.

        Args:
            sample_size (int): the size of the sample. If the size of the sample is greater than
                the size of the CircuitIds then all ids are taken and shuffled.
            inplace (bool): if set to True. Do the transformation inplace.
        """
        sampling = np.random.choice(len(self), size=min(sample_size, len(self)), replace=False)
        return self._apply(lambda x: x[sampling], inplace)

    def limit(self, limit_size, inplace=False):
        """Limit the size of a CircuitIds.

        Notes:
            this function takes the first ``limit_size`` values of a circuit node ids.

        Args:
            limit_size (int): the size of the sample. If the size of the sample is greater than
                the size of the CircuitIds then all ids are kept.
            inplace (bool): if set to True. Do the transformation inplace.
        """
        return self._apply(lambda x: x[:limit_size], inplace)

    def unique(self, inplace=False):
        """Returns only unique CircuitIds.

        Notes:
            this function does not sort the ids
        """
        return self._apply(lambda x: x.unique(), inplace)

    def intersection(self, circuit_ids, inplace=False):
        """Take the intersection of this CircuitIds and the input.

        The index of the resulting object is sorted if ``inplace=False``.
        Otherwise, the orginal order of the index is kept.

        Args:
            circuit_ids (CircuitIds): The CircuitIds to intersect with.
            inplace (bool): if set to True, do the transformation inplace.
        """
        return self._apply(lambda x: x.intersection(circuit_ids.index), inplace)

    def to_csv(self, filepath):
        """Save CircuitIds to csv format."""
        self.index.to_frame(index=False).to_csv(filepath, index=False)

    @classmethod
    def from_csv(cls, filepath):
        """Load CircuitIds from csv."""
        return cls(pd.MultiIndex.from_frame(pd.read_csv(filepath)))

    def tolist(self):
        """Convert the CircuitIds to a list of tuples."""
        return self.index.tolist()

    def __call__(self, *args, **kwargs):
        """Allows to use the CircuitIds as normal indices in a DataFrame."""
        return self.index

    def __len__(self):
        """Return the length of the CircuitIds."""
        return len(self.index)

    def __repr__(self):
        """Correct repr of the CircuitIds."""
        return repr(self.index).replace("MultiIndex", self.__class__.__name__, 1)

    def __eq__(self, other):
        """Equality for the CircuitIds."""
        if self is other:
            return True
        if not isinstance(other, type(self)):
            return False
        return self.index.equals(other.index)

    @abc.abstractmethod
    def __iter__(self):  # pragma: no cover
        """Iterator on the CircuitIds."""

    @abc.abstractmethod
    def __getitem__(self, item):  # pragma: no cover
        """Getter on the CircuitIds."""


class CircuitNodeIds(
    CircuitIds,
    metaclass=AbstractDocSubstitutionMeta,
    source_word="CircuitId",
    target_word="CircuitNodeId",
):
    """High performances CircuitNodeID container."""

    def __init__(self, index, sort_index=True):
        """Return an instance of CircuitNodeIds.

        Args:
            index (pandas.MultiIndex): a multi index from pandas with the population names as
                first level and node IDs as the second level.
            sort_index (bool): if true sort the index per population and then node IDs.
        """
        super().__init__(index, sort_index)
        self.index.names = ["population", "node_ids"]

    def __iter__(self):
        """Iterator on the CircuitNodeIds."""
        for index in self.index:
            yield CircuitNodeId(*index)

    def __getitem__(self, item):
        """Getter on the CircuitNodeIds."""
        return CircuitNodeId(*self.index[item])


class CircuitEdgeIds(
    CircuitIds,
    metaclass=AbstractDocSubstitutionMeta,
    source_word="CircuitId",
    target_word="CircuitEdgeId",
):
    """High performances CircuitEdgeID container."""

    def __init__(self, index, sort_index=True):
        """Return an instance of CircuitEdgeIds.

        Args:
            index (pandas.MultiIndex): a multi index from pandas with the population names as
                first level and node IDs as the second level.
            sort_index (bool): if true sort the index per population and then node IDs.
        """
        super().__init__(index, sort_index)
        self.index.names = ["population", "edge_ids"]

    def __iter__(self):
        """Iterator on the CircuitEdgeIds."""
        for index in self.index:
            yield CircuitEdgeId(*index)

    def __getitem__(self, item):
        """Getter on the CircuitEdgeIds."""
        return CircuitEdgeId(*self.index[item])



# Copyright (c) 2020, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Module containing tools related to the documentation and docstrings."""
import abc
import functools
import inspect
import types


def _word_swapper(doc, source_word, target_word):
    """Swap a word with another in a docstring."""
    if doc is None:
        return doc
    if source_word is None or target_word is None:
        return doc
    return doc.replace(source_word, target_word)


def _copy_func(f):
    """Based on http://stackoverflow.com/a/6528148/190597 (Glenn Maynard)."""
    g = types.FunctionType(
        f.__code__, f.__globals__, name=f.__name__, argdefs=f.__defaults__, closure=f.__closure__
    )
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    return g


class DocSubstitutionMeta(type):
    """Tool to update an inherited class documentation.

    Notes:
        Using a metaclass is better than decorator to do that due to the returned type being
        incorrect when using a wrapper. Ex with CircuitNodeIds:
        type(CircuitNodeIds)
        <class 'bluepysnap._doctools.DocSubstitutionDecorator.__call__.<locals>.Wrapped'>
        when with metaclass:
        type(CircuitNodeIds)
        <class 'bluepysnap.CircuitNodeIds'>
        It works well with Sphinx also.
    """

    def __new__(mcs, name, parents, attrs, source_word=None, target_word=None):
        """Define the new class to return."""
        original_attrs = attrs.copy()
        for parent in parents:
            # skip classmethod with isfunction if I use also ismethod as a predicate I can have the
            # classmethod docstring changed but then the cls argument is not automatically skipped.
            for fun_name, fun_value in inspect.getmembers(parent, predicate=inspect.isfunction):
                # skip abstract methods. This is fine because we must override them anyway
                try:
                    if fun_name in parent.__abstractmethods__:
                        continue
                except AttributeError:
                    pass
                # skip overrode functions
                if fun_name in original_attrs:
                    continue
                # skip special methods
                if fun_name.startswith("__"):
                    continue
                changed_fun = _copy_func(fun_value)
                changed_fun.__doc__ = _word_swapper(changed_fun.__doc__, source_word, target_word)
                attrs[fun_name] = changed_fun
        # create the class
        obj = super(DocSubstitutionMeta, mcs).__new__(mcs, name, parents, attrs)
        return obj


class AbstractDocSubstitutionMeta(abc.ABCMeta, DocSubstitutionMeta):
    """Mixin class to use with abstract classes.

    It solves the metaclass conflict.
    """



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Morphology access."""

from pathlib import Path

import morph_tool.transform as transformations
import numpy as np
from morphio.mut import Morphology

from bluepysnap.exceptions import BluepySnapError
from bluepysnap.sonata_constants import Node
from bluepysnap.utils import is_node_id

EXTENSIONS_MAPPING = {
    "asc": "neurolucida-asc",
    "h5": "h5v1",
}


class MorphHelper:
    """Collection of morphology-related methods."""

    def __init__(self, morph_dir, population, alternate_morphologies=None):
        """Initializes a MorphHelper object from a directory path and a NodePopulation object.

        Args:
            morph_dir (str): Path to the directory containing the node morphologies.
            population (NodePopulation): NodePopulation object used to query the nodes.
            alternate_morphologies (dict): Dictionary containing paths to alternate morphologies.

        Returns:
            MorphHelper: A MorphHelper object.
        """
        self._morph_dir = morph_dir or ""
        self._alternate_morphologies = alternate_morphologies or {}
        self._population = population

    def _get_morph_dir(self, extension):
        """Return morphology directory based on a given extension."""
        if extension == "swc":
            if not self._morph_dir:
                raise BluepySnapError("'morphologies_dir' is not defined in config")
            return self._morph_dir

        alternate_key = EXTENSIONS_MAPPING.get(extension)
        if not alternate_key:
            raise BluepySnapError(f"Unsupported extension: {extension}")

        morph_dir = self._alternate_morphologies.get(alternate_key)
        if not morph_dir:
            raise BluepySnapError(f"'{alternate_key}' is not defined in 'alternate_morphologies'")

        return morph_dir

    def get_filepath(self, node_id, extension="swc"):
        """Return path to SWC morphology file corresponding to `node_id`.

        Args:
            node_id (int/CircuitNodeId): could be a int or CircuitNodeId.
            extension (str): expected filetype extension of the morph file.
        """
        if not is_node_id(node_id):
            raise BluepySnapError("node_id must be a int or a CircuitNodeId")
        name = self._population.get(node_id, Node.MORPHOLOGY)

        return Path(self._get_morph_dir(extension), f"{name}.{extension}")

    def get(self, node_id, transform=False, extension="swc"):
        """Return MorphIO morphology object corresponding to `node_id`.

        Args:
            node_id (int/CircuitNodeId): could be a single int or a CircuitNodeId.
            transform (bool): If `transform` is True, rotate and translate morphology points
                according to `node_id` position in the circuit.
            extension (str): expected filetype extension of the morph file.
        """
        filepath = self.get_filepath(node_id, extension=extension)
        result = Morphology(filepath)
        if transform:
            T = np.eye(4)
            T[:3, :3] = self._population.orientations(node_id)  # rotations
            T[:3, 3] = self._population.positions(node_id).values  # translations
            transformations.transform(result, T)
        return result.as_immutable()



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Edges access."""

import numpy as np

from bluepysnap._doctools import AbstractDocSubstitutionMeta
from bluepysnap.circuit_ids import CircuitEdgeIds, CircuitNodeId, CircuitNodeIds
from bluepysnap.edges import EdgePopulation
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.network import NetworkObject


class Edges(
    NetworkObject,
    metaclass=AbstractDocSubstitutionMeta,
    source_word="NetworkObject",
    target_word="Edge",
):
    """The top level Edges accessor."""

    _population_class = EdgePopulation

    def __init__(self, circuit):  # pylint: disable=useless-super-delegation
        """Initialize the top level Edges accessor."""
        super().__init__(circuit)

    @property
    def population_names(self):
        """Defines all sorted edge population names from the Circuit."""
        return sorted(self._circuit.to_libsonata.edge_populations)

    def ids(self, group=None, sample=None, limit=None):
        """Edge CircuitEdgeIds corresponding to edges ``edge_ids``.

        Args:
            group (None/int/CircuitEdgeId/CircuitEdgeIds/sequence): Which IDs will be
                returned depends on the type of the ``group`` argument:

                - ``None``: return all CircuitEdgeIds.
                - ``CircuitEdgeId``: return the ID in a CircuitEdgeIds object.
                - ``CircuitEdgeIds``: return the IDs in a CircuitNodeIds object.
                - ``int``: returns a CircuitEdgeIds object containing the corresponding edge ID
                  for all populations.
                - ``sequence``: returns a CircuitEdgeIds object containing the corresponding edge
                  IDs for all populations.
            sample (int): If specified, randomly choose ``sample`` number of
                IDs from the match result. If the size of the sample is greater than
                the size of all the EdgePopulations then all ids are taken and shuffled.
            limit (int): If specified, return the first ``limit`` number of
                IDs from the match result. If limit is greater than the size of all the populations
                all node IDs are returned.

        Returns:
            CircuitEdgeIds: returns a CircuitEdgeIds containing all the edge IDs and the
            corresponding populations. For performance reasons we do not test if the edge ids
            are present or not in the circuit.

        Notes:
            This envision also the maybe future selection of edges on queries.
        """
        if isinstance(group, CircuitEdgeIds):
            diff = np.setdiff1d(group.get_populations(unique=True), self.population_names)
            if diff.size != 0:
                raise BluepySnapError(f"Population {diff} does not exist in the circuit.")
        fun = lambda x: (x.ids(group), x.name)
        return self._get_ids_from_pop(fun, CircuitEdgeIds, sample=sample, limit=limit)

    def get(self, edge_ids=None, properties=None):  # pylint: disable=arguments-renamed
        """Edge properties as pandas DataFrame.

        Args:
            edge_ids (int/CircuitEdgeId/CircuitEdgeIds/sequence): same as Edges.ids().
            properties (None/str/list): an edge property name or a list of edge property names.
                If set to None ids are returned.

        Returns:
            pandas.Series/pandas.DataFrame:
                - A pandas Series indexed by edge IDs if ``properties`` is scalar.
                - A pandas DataFrame indexed by edge IDs if ``properties`` is list.

        Notes:
            The Edges.property_names function will give you all the usable properties
            for the `properties` argument.
        """
        if edge_ids is None:
            raise BluepySnapError("You need to set edge_ids in get.")
        if properties is None:
            return edge_ids
        return super().get(edge_ids, properties)

    def afferent_nodes(self, target, unique=True):
        """Get afferent CircuitNodeIDs for given target ``node_id``.

        Notes:
            Afferent nodes are nodes projecting an outgoing edge to one of the ``target`` node.

        Args:
            target (CircuitNodeIds/int/sequence/str/mapping/None): the target you want to resolve
            and use as target nodes.
            unique (bool): If ``True``, return only unique afferent node IDs.

        Returns:
            CircuitNodeIDs: Afferent CircuitNodeIDs for all the targets from all edge population.
        """
        target_ids = self._circuit.nodes.ids(target)
        result = self._get_ids_from_pop(
            lambda x: (x.afferent_nodes(target_ids), x.source.name), CircuitNodeIds
        )
        if unique:
            result.unique(inplace=True)
        return result

    def efferent_nodes(self, source, unique=True):
        """Get efferent node IDs for given source ``node_id``.

        Notes:
            Efferent nodes are nodes receiving an incoming edge from one of the ``source`` node.

        Args:
            source (CircuitNodeIds/int/sequence/str/mapping/None): the source you want to resolve
                and use as source nodes.
            unique (bool): If ``True``, return only unique afferent node IDs.

        Returns:
            numpy.ndarray: Efferent node IDs for all the sources.
        """
        source_ids = self._circuit.nodes.ids(source)
        result = self._get_ids_from_pop(
            lambda x: (x.efferent_nodes(source_ids), x.target.name), CircuitNodeIds
        )
        if unique:
            result.unique(inplace=True)
        return result

    def pathway_edges(self, source=None, target=None, properties=None):
        """Get edges corresponding to ``source`` -> ``target`` connections.

        Args:
            source: source node group
            target: target node group
            properties: None / edge property name / list of edge property names

        Returns:
            - CircuitEdgeIDs, if ``properties`` is None;
            - Pandas Series indexed by CircuitEdgeIDs if ``properties`` is string;
            - Pandas DataFrame indexed by CircuitEdgeIDs if ``properties`` is list.
        """
        if source is None and target is None:
            raise BluepySnapError("Either `source` or `target` should be specified")

        source_ids = self._circuit.nodes.ids(source)
        target_ids = self._circuit.nodes.ids(target)

        result = self._get_ids_from_pop(
            lambda x: (x.pathway_edges(source_ids, target_ids), x.name), CircuitEdgeIds
        )

        if properties:
            return self.get(result, properties)
        return result

    def afferent_edges(self, node_id, properties=None):
        """Get afferent edges for given ``node_id``.

        Args:
            node_id (int): Target node ID.
            properties: An edge property name, a list of edge property names, or None.

        Returns:
            pandas.Series/pandas.DataFrame/list:
                - A pandas Series indexed by edge ID if ``properties`` is a string.
                - A pandas DataFrame indexed by edge ID if ``properties`` is a list.
                - A list of edge IDs, if ``properties`` is None.
        """
        return self.pathway_edges(source=None, target=node_id, properties=properties)

    def efferent_edges(self, node_id, properties=None):
        """Get efferent edges for given ``node_id``.

        Args:
            node_id: source node ID
            properties: None / edge property name / list of edge property names

        Returns:
            - List of edge IDs, if ``properties`` is None;
            - Pandas Series indexed by edge IDs if ``properties`` is string;
            - Pandas DataFrame indexed by edge IDs if ``properties`` is list.
        """
        return self.pathway_edges(source=node_id, target=None, properties=properties)

    def pair_edges(self, source_node_id, target_node_id, properties=None):
        """Get edges corresponding to ``source_node_id`` -> ``target_node_id`` connection.

        Args:
            source_node_id: source node ID
            target_node_id: target node ID
            properties: None / edge property name / list of edge property names

        Returns:
            - List of edge IDs, if ``properties`` is None;
            - Pandas Series indexed by edge IDs if ``properties`` is string;
            - Pandas DataFrame indexed by edge IDs if ``properties`` is list.
        """
        return self.pathway_edges(
            source=source_node_id, target=target_node_id, properties=properties
        )

    @staticmethod
    def _add_circuit_ids(its, source, target):
        """Generator comprehension adding the CircuitIds to the iterator.

        Notes:
            Using closures or lambda functions would result in override functions and so the
            source and target would be the same for all the populations.
        """
        return (
            (CircuitNodeId(source, source_id), CircuitNodeId(target, target_id), count)
            for source_id, target_id, count in its
        )

    @staticmethod
    def _add_edge_ids(its, source, target, pop_name):
        """Generator comprehension adding the CircuitIds to the iterator."""
        return (
            (
                CircuitNodeId(source, source_id),
                CircuitNodeId(target, target_id),
                CircuitEdgeIds.from_dict({pop_name: edge_id}),
            )
            for source_id, target_id, edge_id in its
        )

    @staticmethod
    def _omit_edge_count(its, source, target):
        """Generator comprehension adding the CircuitIds to the iterator."""
        return (
            (CircuitNodeId(source, source_id), CircuitNodeId(target, target_id))
            for source_id, target_id in its
        )

    def iter_connections(
        self, source=None, target=None, return_edge_ids=False, return_edge_count=False
    ):
        """Iterate through ``source`` -> ``target`` connections.

        Args:
            source (CircuitNodeIds/int/sequence/str/mapping/None): source node group
            target (CircuitNodeIds/int/sequence/str/mapping/None): target node group
            return_edge_count: if True, edge count is added to yield result
            return_edge_ids: if True, edge ID list is added to yield result

        ``return_edge_count`` and ``return_edge_ids`` are mutually exclusive.

        Yields:
            - (source_node_id, target_node_id, edge_ids) if ``return_edge_ids`` is True;
            - (source_node_id, target_node_id, edge_count) if ``return_edge_count`` is True;
            - (source_node_id, target_node_id) otherwise.
        """
        if return_edge_ids and return_edge_count:
            raise BluepySnapError(
                "`return_edge_count` and `return_edge_ids` are mutually exclusive"
            )
        for name, pop in self.items():
            it = pop.iter_connections(
                source=source,
                target=target,
                return_edge_ids=return_edge_ids,
                return_edge_count=return_edge_count,
            )
            source_pop = pop.source.name
            target_pop = pop.target.name
            if return_edge_count:
                yield from self._add_circuit_ids(it, source_pop, target_pop)
            elif return_edge_ids:
                yield from self._add_edge_ids(it, source_pop, target_pop, name)
            else:
                yield from self._omit_edge_count(it, source_pop, target_pop)

    def __getstate__(self):
        """Make Edges pickle-able, without storing state of caches."""
        return self._circuit

    def __setstate__(self, state):
        """Load from pickle state."""
        self.__init__(state)



import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest

from bluepysnap import BluepySnapError
from bluepysnap.query import _circuit_mask, _logical_and, _logical_or, _positional_mask, resolve_ids


def test_positional_mask():
    data = pd.DataFrame(range(3))
    npt.assert_array_equal(_positional_mask(data, [1, 2]), [False, True, True])
    npt.assert_array_equal(_positional_mask(data, [0, 2]), [True, False, True])


def test_population_mask():
    data = pd.DataFrame(range(3))
    queries, mask = _circuit_mask(data, "default", {"population": "default", "other": "val"})
    assert queries == {"other": "val"}
    npt.assert_array_equal(mask, [True, True, True])

    queries, mask = _circuit_mask(data, "default", {"population": "unknown", "other": "val"})
    assert queries == {"other": "val"}
    npt.assert_array_equal(mask, [False, False, False])

    queries, mask = _circuit_mask(
        data, "default", {"population": "default", "node_id": [2], "other": "val"}
    )
    assert queries == {"other": "val"}
    npt.assert_array_equal(mask, [False, False, True])

    queries, mask = _circuit_mask(data, "default", {"other": "val"})
    assert queries == {"other": "val"}
    npt.assert_array_equal(mask, [True, True, True])


def test_resolve_ids():
    data = pd.DataFrame(
        [[1, 0.4, "seven"], [2, 0.5, "eight"], [3, 0.6, "nine"]], columns=["int", "float", "str"]
    )
    assert [False, True, False] == resolve_ids(data, "", {"str": "eight"}).tolist()
    assert [False, False, False] == resolve_ids(data, "", {"int": 1, "str": "eight"}).tolist()
    assert [True, True, False] == resolve_ids(
        data, "", {"$or": [{"str": "seven"}, {"float": (0.5, 0.5)}]}
    ).tolist()
    assert [False, False, True] == resolve_ids(
        data, "", {"$and": [{"str": "nine"}, {"int": 3}]}
    ).tolist()
    assert [False, False, True] == resolve_ids(
        data, "", {"$and": [{"str": "nine"}, {"$or": [{"int": 1}, {"int": 3}]}]}
    ).tolist()
    assert [False, True, True] == resolve_ids(
        data, "", {"$or": [{"float": (0.59, 0.61)}, {"$and": [{"str": "eight"}, {"int": 2}]}]}
    ).tolist()
    assert [True, False, True] == resolve_ids(
        data, "", {"$or": [{"node_id": 0}, {"edge_id": 2}]}
    ).tolist()
    assert [False, False, False] == resolve_ids(data, "", {"$or": []}).tolist()
    assert [True, True, True] == resolve_ids(data, "", {"$and": []}).tolist()

    with pytest.raises(BluepySnapError) as e:
        resolve_ids(data, "", {"str": {"$regex": "*.some", "edge_id": 2}})
    assert "Value operators can't be used with plain values" in e.value.args[0]


@pytest.mark.parametrize(
    "masks, expected",
    [
        ([], True),
        ([True], True),
        ([False], False),
        ([True, False], False),
        ([np.array([True])], np.array([True])),
        ([np.array([False])], np.array([False])),
        ([np.array([True, True]), np.array([True, False])], np.array([True, False])),
        ([np.array([100, 100]), np.array([True, False])], np.array([True, False])),
        ([pd.Series([100, 100]), np.array([True, False])], np.array([True, False])),
        (
            [np.array([True, False, True, False]), np.array([True, True, False, False])],
            np.array([True, False, False, False]),
        ),
        ([np.array([1, 0, 1, 0]), np.array([1, 1, 0, 0])], np.array([True, False, False, False])),
        (
            [np.array([1, 0, 1, 0]), np.array([1, 1, 0, 0]), True],
            np.array([True, False, False, False]),
        ),
        ([np.array([1, 0, 1, 0]), np.array([1, 1, 0, 0]), False], False),
        (
            [
                np.array([1, 0, 1, 0, 1, 0, 1, 0]),
                np.array([1, 1, 0, 0, 1, 1, 0, 0]),
                np.array([1, 1, 1, 1, 0, 0, 0, 0]),
            ],
            np.array([True, False, False, False, False, False, False, False]),
        ),
    ],
)
def test__logical_and(masks, expected):
    result = _logical_and(masks)

    assert type(result) is type(expected)
    npt.assert_equal(result, expected)


@pytest.mark.parametrize(
    "masks, expected",
    [
        ([], False),
        ([True], True),
        ([False], False),
        ([True, False], True),
        ([np.array([True])], np.array([True])),
        ([np.array([False])], np.array([False])),
        ([np.array([True, True]), np.array([True, False])], np.array([True, True])),
        ([np.array([100, 100]), np.array([True, False])], np.array([True, True])),
        ([pd.Series([100, 100]), np.array([True, False])], np.array([True, True])),
        (
            [np.array([True, False, True, False]), np.array([True, True, False, False])],
            np.array([True, True, True, False]),
        ),
        ([np.array([1, 0, 1, 0]), np.array([1, 1, 0, 0])], np.array([True, True, True, False])),
        (
            [np.array([1, 0, 1, 0]), np.array([1, 1, 0, 0]), False],
            np.array([True, True, True, False]),
        ),
        ([np.array([1, 0, 1, 0]), np.array([1, 1, 0, 0]), True], True),
        (
            [
                np.array([1, 0, 1, 0, 1, 0, 1, 0]),
                np.array([1, 1, 0, 0, 1, 1, 0, 0]),
                np.array([1, 1, 1, 1, 0, 0, 0, 0]),
            ],
            np.array([True, True, True, True, True, True, True, False]),
        ),
    ],
)
def test__logical_or(masks, expected):
    result = _logical_or(masks)

    assert type(result) is type(expected)
    npt.assert_equal(result, expected)



"""Functions for schema based validation of circuit files."""
from pathlib import Path

import h5py
import importlib_resources
import jsonschema
import numpy as np
import referencing
import yaml

DEFINITIONS = "definitions"


def _load_schema_file(*args):
    """Load one of the predefined YAML schema files."""
    filename = str(Path(*args).with_suffix(".yaml"))
    filepath = importlib_resources.files(__package__) / filename
    if not filepath.is_file():
        raise FileNotFoundError(f"Schema file {filepath} not found")
    return yaml.safe_load(filepath.read_text())


def _parse_path(path, join_str):
    error_path = []

    for item in path:
        # show list index in the [0] format
        if isinstance(item, int):
            error_path[-1] += f"[{item}]"
        else:
            error_path.append(item)

    return join_str.join(error_path)


def _wrap_errors(filepath, schema_errors, join_str):
    """Handles parsing of schema errors into more meaningful messages.

    Also wraps all the warngings and errors to single Error instances.

    join_str allows showing circuit config paths as:
        path.to.the.error[1].cause
    and h5 dataset paths as:
        path/to/the/error/cause
    """
    # NOTE: would probably make more sense to have different parser for circuit,
    # as datatype and attributes only exists in the case of h5 files.
    from bluepysnap.circuit_validation import Error

    warnings = []
    errors = []

    for e in schema_errors:
        if not e.path:
            errors.append(e.message)
        elif e.path[-1] == "datatype":
            path = _parse_path(list(e.path)[:-1], join_str)
            warnings.append(f"incorrect datatype '{e.instance}' for '{path}': {e.message}")
        else:
            if e.schema_path[-1] in ("maxProperties", "minProperties"):
                path = _parse_path(e.path, join_str)
                many_or_few = "many" if e.schema_path[-1] == "maxProperties" else "few"
                message = f"{path}: too {many_or_few} properties"
            elif e.path[-1] == "attributes":
                # Add special message in case of attribute missing
                path = _parse_path(list(e.path)[:-1], join_str)
                attributes = e.schema["required"]
                message = f"{path}: missing required attribute(s) {attributes}"
            elif "attributes" in e.schema.get("required", []):
                path = _parse_path(e.path, join_str)
                attributes = e.schema["properties"]["attributes"]["required"]
                message = f"{path}: missing required attribute(s) {attributes}"
            else:
                path = _parse_path(e.path, join_str)
                message = f"{path}: {e.message}"
            errors.append(message)

    ret_errors = []

    if len(warnings) > 0:
        message = filepath + ":\n\t" + "\n\t".join(warnings)
        ret_errors.append(Error(Error.WARNING, message))
    if len(errors) > 0:
        message = filepath + ":\n\t" + "\n\t".join(errors)
        ret_errors.append(Error(Error.FATAL, message))

    return ret_errors


def _validate_schema_for_dict(schema, dict_):
    """Run a schema validation for a dictionary."""
    validator = jsonschema.validators.Draft202012Validator(schema)

    return validator.iter_errors(dict_)


def _parse_schema(object_type, sub_type=None):
    """Parses the schema from partial schemas for given object type.

    Args:
        object_type (str): Type of the object. Accepts "edge", "node" and "circuit".
        sub_type (str): Sub type of object. E.g., "biophysical".

    Returns:
        dict: Schema parsed as a dictionary.
    """
    if object_type == "circuit":
        return _load_schema_file(object_type)
    elif object_type not in ("edge", "node"):
        raise RuntimeError(f"Unknown object type: {object_type}")

    schema = _load_schema_file(DEFINITIONS, "datatypes")
    schema.update(_load_schema_file(DEFINITIONS, object_type))
    schema.update(_load_schema_file(object_type, sub_type))

    return schema


def _get_h5_structure_as_dict(h5):
    """Recursively translates h5 file into a dictionary.

    For groups, the subgroups/datasets are translated as a subdictionary.
    Datatype of datasets is resolved and returned as {'datatype': <dtype>}.
    Attributes of either groups or datasets are returned as {'attributes': {<key>: <value>}}.

    Args:
        h5 (h5.File, h5.Group): h5 file or group to translate

    Returns:
        dict: dictionary of the structure
    """
    properties = {}

    def get_dataset_dtype(item):
        if item.dtype.hasobject:
            return h5py.check_string_dtype(item.dtype).encoding

        return item.dtype.name

    for key, value in h5.items():
        if isinstance(value, h5py.Group):
            properties[key] = _get_h5_structure_as_dict(value)
        else:  # Dataset
            properties[key] = {"datatype": get_dataset_dtype(value)}

        attrs = dict(value.attrs.items())
        if attrs:
            properties[key]["attributes"] = attrs

    # Resolve enumerated properties' types to check their actual data type
    if "@library" in properties:
        for key, value in properties["@library"].items():
            properties[key] = value

    return properties


def validate_circuit_schema(path, config):
    """Validates a circuit config against a schema.

    Args:
        path (str): path to the config (for error messages)
        config (dict): resolved bluepysnap config

    Returns:
        list: List of errors, empty if no errors
    """
    errors = _validate_schema_for_dict(_parse_schema("circuit"), config)

    return _wrap_errors(path, errors, ".")


def validate_nodes_schema(path, nodes_type):
    """Validates a nodes file against a schema.

    Args:
        path (str): path to the nodes file
        nodes_type (str): node type (e.g., "biophysical")

    Returns:
        list: List of errors, empty if no errors
    """
    with h5py.File(path) as h5:
        nodes_h5_dict = _get_h5_structure_as_dict(h5)

    errors = _validate_schema_for_dict(_parse_schema("node", nodes_type), nodes_h5_dict)

    return _wrap_errors(path, errors, "/")


def validate_edges_schema(path, edges_type, virtual):
    """Validates an edges file against a schema.

    Args:
        path (str): path to the edges file
        edges_type (str): edge type (e.g., "chemical")
        virtual(bool): whether this is a virtual edge population

    Returns:
        list: List of errors, empty if no errors
    """
    if virtual:
        edges_type += "_virtual"

    with h5py.File(path) as h5:
        edges_h5_dict = _get_h5_structure_as_dict(h5)

    errors = _validate_schema_for_dict(_parse_schema("edge", edges_type), edges_h5_dict)

    return _wrap_errors(path, errors, "/")


def _resolve_types(resolver, types):
    """Use jsonschema `resolver` to resolve the `types` dict."""
    cache = {}

    def _resolve_type(type_):
        if type_ not in cache:
            type_ = resolver.lookup(type_).contents["properties"]["datatype"]["const"]
            if hasattr(np, type_):
                cache[type_] = getattr(np, type_)
            elif type_ == "utf-8":
                cache[type_] = str

        return cache[type_]

    return {k: _resolve_type(v["$ref"]) for k, v in types.items()}


def _get_reference_resolver(schema):
    """Get reference resolver for the given schema."""
    resource = referencing.Resource(
        contents=schema,
        specification=referencing.jsonschema.DRAFT202012,
    )
    return referencing.Registry().with_resource("", resource=resource).resolver()


def nodes_schema_types(nodes_type):
    """Get the datatypes of the attributes for nodes.

    Args:
        nodes_type (str): node type (e.g., "biophysical")

    Returns:
        dict: name -> type of column
    """
    schema = _parse_schema("node", nodes_type)
    resolver = _get_reference_resolver(schema)

    schema = schema["$node_file_defs"]["nodes_file_root"]["properties"]["nodes"]
    schema = schema["patternProperties"][""]["properties"]["0"]["properties"]
    dynamics_params = schema["dynamics_params"]["properties"]
    del schema["dynamics_params"]
    del schema["@library"]

    return _resolve_types(resolver, schema), _resolve_types(resolver, dynamics_params)


def edges_schema_types(edges_type, virtual):
    """Get the datatypes of the attributes for edges.

    Args:
        edges_type (str): edges type (e.g., "chemical")
        virtual(bool): whether this is a virtual edge population

    Returns:
        dict: name -> type of column
    """
    if virtual:
        edges_type += "_virtual"

    schema = _parse_schema("edge", edges_type)
    resolver = _get_reference_resolver(schema)

    schema = schema["$edge_file_defs"]["edges_file_root"]["properties"]["edges"]
    schema = schema["patternProperties"][""]["properties"]["0"]["properties"]
    del schema["@library"]
    del schema["synapse_id"]

    return _resolve_types(resolver, schema)



import json
import pickle

import pandas as pd
import pytest
from libsonata import SonataError

import bluepysnap.circuit as test_module
from bluepysnap.edges import EdgePopulation, Edges
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.nodes import NodePopulation, Nodes

from utils import PICKLED_SIZE_ADJUSTMENT, TEST_DATA_DIR, copy_test_data, edit_config


def test_all():
    circuit = test_module.Circuit(str(TEST_DATA_DIR / "circuit_config.json"))
    assert circuit.config["networks"]["nodes"][0] == {
        "nodes_file": str(TEST_DATA_DIR / "nodes.h5"),
        "populations": {
            "default": {"type": "biophysical"},
            "default2": {"type": "biophysical", "spatial_segment_index_dir": "path/to/node/dir"},
        },
    }
    assert isinstance(circuit.nodes, Nodes)
    assert isinstance(circuit.edges, Edges)
    assert list(circuit.edges) == ["default", "default2"]
    assert isinstance(circuit.edges["default"], EdgePopulation)
    assert isinstance(circuit.edges["default2"], EdgePopulation)
    assert sorted(list(circuit.nodes)) == ["default", "default2"]
    assert isinstance(circuit.nodes["default"], NodePopulation)
    assert isinstance(circuit.nodes["default2"], NodePopulation)
    assert sorted(circuit.node_sets) == sorted(
        json.loads((TEST_DATA_DIR / "node_sets.json").read_text())
    )

    fake_pop = "fake"
    with pytest.raises(
        BluepySnapError, match=f"Population config not found for node population: {fake_pop}"
    ):
        circuit.get_node_population_config(fake_pop)
    with pytest.raises(
        BluepySnapError, match=f"Population config not found for edge population: {fake_pop}"
    ):
        circuit.get_edge_population_config(fake_pop)


def test_duplicate_node_populations():
    with copy_test_data() as (_, config_path):
        with edit_config(config_path) as config:
            config["networks"]["nodes"].append(config["networks"]["nodes"][0])

        match = "Duplicate population|Population default is declared twice"
        with pytest.raises(SonataError, match=match):
            test_module.Circuit(config_path)


def test_duplicate_edge_populations():
    with copy_test_data() as (_, config_path):
        with edit_config(config_path) as config:
            config["networks"]["edges"].append(config["networks"]["edges"][0])

        match = "Duplicate population|Population default is declared twice"
        with pytest.raises(SonataError, match=match):
            test_module.Circuit(config_path)


def test_no_node_set():
    with copy_test_data() as (_, config_path):
        with edit_config(config_path) as config:
            config.pop("node_sets_file")
        circuit = test_module.Circuit(config_path)
        assert circuit.node_sets.content == {}


def test_integration():
    circuit = test_module.Circuit(str(TEST_DATA_DIR / "circuit_config.json"))
    node_ids = circuit.nodes.ids({"mtype": ["L6_Y", "L2_X"]})
    edge_ids = circuit.edges.afferent_edges(node_ids)
    edge_props = circuit.edges.get(edge_ids, properties=["syn_weight", "delay"])
    edge_reduced = edge_ids.limit(2)
    edge_props = pd.concat(df for _, df in edge_props)
    edge_props_reduced = edge_props.loc[edge_reduced]
    assert edge_props_reduced["syn_weight"].tolist() == [1, 1]


def test_pickle(tmp_path):
    circuit = test_module.Circuit(TEST_DATA_DIR / "circuit_config.json")

    pickle_path = tmp_path / "pickle.pkl"
    with open(pickle_path, "wb") as fd:
        pickle.dump(circuit, fd)

    with open(pickle_path, "rb") as fd:
        circuit = pickle.load(fd)

    assert pickle_path.stat().st_size < 60 + PICKLED_SIZE_ADJUSTMENT
    assert list(circuit.edges) == ["default", "default2"]


def test_empty_manifest(tmp_path):
    path = tmp_path / "circuit_config.json"
    with open(path, "w") as fd:
        json.dump(
            {
                "manifest": {},
                "networks": {
                    "nodes": {},
                    "edges": {},
                },
            },
            fd,
        )

    test_module.Circuit(path)



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Nodes access."""

import numpy as np

from bluepysnap._doctools import AbstractDocSubstitutionMeta
from bluepysnap.circuit_ids import CircuitNodeIds
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.network import NetworkObject
from bluepysnap.nodes import NodePopulation


class Nodes(
    NetworkObject,
    metaclass=AbstractDocSubstitutionMeta,
    source_word="NetworkObject",
    target_word="Node",
):
    """The top level Nodes accessor."""

    _population_class = NodePopulation

    def __init__(self, circuit):  # pylint: disable=useless-super-delegation
        """Initialize the top level Nodes accessor."""
        super().__init__(circuit)

    @property
    def population_names(self):
        """Defines all sorted node population names from the Circuit."""
        return sorted(self._circuit.to_libsonata.node_populations)

    def property_values(self, prop):
        """Returns all the values for a given Nodes property."""
        return set(
            value
            for pop in self.values()
            if prop in pop.property_names
            for value in pop.property_values(prop)
        )

    def ids(self, group=None, sample=None, limit=None):
        """Returns the CircuitNodeIds corresponding to the nodes from ``group``.

        Args:
            group (CircuitNodeId/CircuitNodeIds/int/sequence/str/mapping/None): Which IDs will be
                returned depends on the type of the ``group`` argument:

                - ``CircuitNodeId``: return the ID in a CircuitNodeIds object if it belongs to
                  the circuit.
                - ``CircuitNodeIds``: return the IDs in a CircuitNodeIds object if they belong to
                  the circuit.
                - ``int``: if the node ID is present in all populations, returns a CircuitNodeIds
                  object containing the corresponding node ID for all populations.
                - ``sequence``: if all the values contained in the sequence are present in all
                  populations, returns a CircuitNodeIds object containing the corresponding node
                  IDs for all populations.
                - ``str``: use a node set name as input. Returns a CircuitNodeIds object containing
                  nodes selected by the node set.
                - ``mapping``: Returns a CircuitNodeIds object containing nodes matching a
                  properties filter.
                - ``None``: return all node IDs of the circuit in a CircuitNodeIds object.
            sample (int): If specified, randomly choose ``sample`` number of
                IDs from the match result. If the size of the sample is greater than
                the size of all the NodePopulations then all ids are taken and shuffled.
            limit (int): If specified, return the first ``limit`` number of
                IDs from the match result. If limit is greater than the size of all the populations,
                all node IDs are returned.

        Returns:
            CircuitNodeIds: returns a CircuitNodeIds containing all the node IDs and the
            corresponding populations. All the explicitly requested IDs must be present inside
            the circuit.

        Raises:
            BluepySnapError: when a population from a CircuitNodeIds is not present in the circuit.
            BluepySnapError: when an id query via a int, sequence, or CircuitNodeIds is not present
                in the circuit.

        Examples:
            The available group parameter values (example with 2 node populations pop1 and pop2):

            >>> nodes = circuit.nodes
            >>> nodes.ids(group=None)  #  returns all CircuitNodeIds from the circuit
            >>> node_ids = CircuitNodeIds.from_arrays(["pop1", "pop2"], [1, 3])
            >>> nodes.ids(group=node_ids)  #  returns ID 1 from pop1 and ID 3 from pop2
            >>> nodes.ids(group=0)  #  returns CircuitNodeIds 0 from pop1 and pop2
            >>> nodes.ids(group=[0, 1])  #  returns CircuitNodeIds 0 and 1 from pop1 and pop2
            >>> nodes.ids(group="node_set_name")  # returns CircuitNodeIds matching node set
            >>> nodes.ids(group={Node.LAYER: 2})  # returns CircuitNodeIds matching layer==2
            >>> nodes.ids(group={Node.LAYER: [2, 3]})  # returns CircuitNodeIds with layer in [2,3]
            >>> nodes.ids(group={Node.X: (0, 1)})  # returns CircuitNodeIds with 0 < x < 1
            >>> # returns CircuitNodeIds matching one of the queries inside the 'or' list
            >>> nodes.ids(group={'$or': [{ Node.LAYER: [2, 3]},
            >>>                          { Node.X: (0, 1), Node.MTYPE: 'L1_SLAC' }]})
            >>> # returns CircuitNodeIds matching all the queries inside the 'and' list
            >>> nodes.ids(group={'$and': [{ Node.LAYER: [2, 3]},
            >>>                           { Node.X: (0, 1), Node.MTYPE: 'L1_SLAC' }]})
        """
        if isinstance(group, CircuitNodeIds):
            diff = np.setdiff1d(group.get_populations(unique=True), self.population_names)
            if diff.size != 0:
                raise BluepySnapError(f"Population {diff} does not exist in the circuit.")

        fun = lambda x: (x.ids(group, raise_missing_property=False), x.name)
        return self._get_ids_from_pop(fun, CircuitNodeIds, sample=sample, limit=limit)

    def get(self, group=None, properties=None):  # pylint: disable=arguments-differ
        """Node properties as a pandas DataFrame.

        Args:
            group (CircuitNodeIds/int/sequence/str/mapping/None): Which nodes will have their
                properties returned depends on the type of the ``group`` argument:
                See :py:class:`~bluepysnap.nodes.Nodes.ids`.

            properties (str/list): If specified, return only the properties in the list.
                Otherwise return all properties.

        Returns:
            pandas.DataFrame: Return a pandas DataFrame indexed by NodeCircuitIds containing the
                properties from ``properties``.

        Notes:
            The NodePopulation.property_names function will give you all the usable properties
            for the `properties` argument.
        """
        if properties is None:
            # not strictly needed, but ensure that the properties are always in the same order
            properties = sorted(self.property_names)
        return super().get(group, properties)

    def __getstate__(self):
        """Make Nodes pickle-able, without storing state of caches."""
        return self._circuit

    def __setstate__(self, state):
        """Load from pickle state."""
        self.__init__(state)



from unittest.mock import ANY, Mock

import numpy as np
import numpy.testing as npt
import pytest

import bluepysnap.edges.edge_population as test_module


def test_estimate_range_size_1():
    func = lambda x: Mock(ranges=np.zeros(x))
    actual = test_module._estimate_range_size(func, [11, 21, 31], n=5)
    npt.assert_equal(actual, 21)


def test_estimate_range_size_2():
    func = lambda x: Mock(ranges=[42])
    actual = test_module._estimate_range_size(func, range(10))
    npt.assert_equal(actual, 1)


def test_estimate_range_size_3():
    func = lambda x: Mock(ranges=[42])
    actual = test_module._estimate_range_size(func, range(10))
    npt.assert_equal(actual, 1)


def test_estimate_range_size_4():
    with pytest.raises(AssertionError):
        test_module._estimate_range_size(ANY, [])



import pickle
import sys
from pathlib import Path
from unittest import mock

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest

import bluepysnap.edges.edge_population as test_module
from bluepysnap.bbp import Synapse
from bluepysnap.circuit import Circuit
from bluepysnap.circuit_ids import CircuitEdgeId, CircuitEdgeIds, CircuitNodeIds
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.sonata_constants import DEFAULT_EDGE_TYPE, Edge
from bluepysnap.utils import IDS_DTYPE

from utils import PICKLED_SIZE_ADJUSTMENT, TEST_DATA_DIR, copy_test_data, edit_config


def index_as_ids_dtypes(values):
    """have pandas index types match"""
    return np.array(values, dtype=IDS_DTYPE)


class TestEdgePopulation:
    @staticmethod
    def get_edge_population(circuit_path, pop_name):
        circuit = Circuit(circuit_path)
        return test_module.EdgePopulation(circuit, pop_name)

    def setup_method(self):
        self.test_obj = TestEdgePopulation.get_edge_population(
            TEST_DATA_DIR / "circuit_config.json", "default"
        )

    def test_basic(self):
        assert self.test_obj.name == "default"
        assert self.test_obj.source.name == "default"
        assert self.test_obj.target.name == "default"
        assert self.test_obj.size, 4
        assert sorted(self.test_obj.property_names) == sorted(
            [
                Synapse.SOURCE_NODE_ID,
                Synapse.TARGET_NODE_ID,
                Synapse.AXONAL_DELAY,
                Synapse.G_SYNX,
                Synapse.POST_X_CENTER,
                Synapse.POST_Y_CENTER,
                Synapse.POST_Z_CENTER,
                Synapse.POST_X_SURFACE,
                Synapse.POST_Y_SURFACE,
                Synapse.POST_Z_SURFACE,
                Synapse.PRE_X_CENTER,
                Synapse.PRE_Y_CENTER,
                Synapse.PRE_Z_CENTER,
                Synapse.PRE_X_SURFACE,
                Synapse.PRE_Y_SURFACE,
                Synapse.PRE_Z_SURFACE,
                Synapse.POST_SECTION_ID,
                Synapse.POST_SECTION_POS,
                Synapse.PRE_SECTION_ID,
                Synapse.PRE_SECTION_POS,
                Synapse.SYN_WEIGHT,
                test_module.DYNAMICS_PREFIX + "param1",
            ]
        )
        assert self.test_obj.type == DEFAULT_EDGE_TYPE

    def test_population_type(self):
        with copy_test_data() as (config_dir, config_path):
            with edit_config(config_path) as config:
                config["networks"]["edges"] = [
                    {
                        "edge_types_file": None,
                        "edges_file": str(Path(config_dir) / "edges_complete_graph.h5"),
                        "populations": {"default": {"type": "fake_type"}},
                    }
                ]

            test_obj = TestEdgePopulation.get_edge_population(config_path, "default")

            assert test_obj.type == "fake_type"

    def test_container_properties(self):
        expected = sorted(
            [
                "PRE_Y_SURFACE",
                "PRE_Z_SURFACE",
                "PRE_X_CENTER",
                "POST_Y_CENTER",
                "AXONAL_DELAY",
                "POST_X_CENTER",
                "POST_Y_SURFACE",
                "POST_Z_SURFACE",
                "PRE_Y_CENTER",
                "POST_Z_CENTER",
                "PRE_Z_CENTER",
                "PRE_X_SURFACE",
                "POST_X_SURFACE",
                "POST_SECTION_ID",
                "PRE_SECTION_ID",
                "POST_SECTION_POS",
                "PRE_SECTION_POS",
                "SYN_WEIGHT",
                "SOURCE_NODE_ID",
                "TARGET_NODE_ID",
            ]
        )
        assert sorted(self.test_obj.container_property_names(Edge)) == expected
        with pytest.raises(BluepySnapError):
            mapping = {"X": "x"}
            self.test_obj.container_property_names(mapping)

        with pytest.raises(BluepySnapError):
            self.test_obj.container_property_names(int)

    def test_nodes(self):
        assert self.test_obj._nodes("default").name == "default"
        with pytest.raises(BluepySnapError):
            self.test_obj._nodes("no-such-population")

    def test_property_dtypes(self):
        from numpy import dtype

        expected = pd.Series(
            data=[
                dtype("float32"),
                dtype("float64"),
                dtype("float64"),
                dtype("float64"),
                dtype("float32"),
                dtype("float64"),
                dtype("float32"),
                dtype("float64"),
                dtype("int64"),
                dtype("int64"),
                dtype("float64"),
                dtype("float64"),
                dtype("float64"),
                dtype("float64"),
                dtype("float64"),
                dtype("float64"),
                dtype("float32"),
                dtype("float32"),
                dtype("float64"),
                dtype("float64"),
                dtype("int64"),
                dtype("int64"),
            ],
            index=[
                "syn_weight",
                "@dynamics:param1",
                "afferent_surface_y",
                "afferent_surface_z",
                "conductance",
                "efferent_center_x",
                "delay",
                "afferent_center_z",
                "efferent_section_id",
                "afferent_section_id",
                "efferent_center_y",
                "afferent_center_x",
                "efferent_surface_z",
                "afferent_center_y",
                "afferent_surface_x",
                "efferent_surface_x",
                "afferent_section_pos",
                "efferent_section_pos",
                "efferent_surface_y",
                "efferent_center_z",
                "@source_node",
                "@target_node",
            ],
        ).sort_index()

        pdt.assert_series_equal(expected, self.test_obj.property_dtypes)

    def test_ids(self):
        tested = self.test_obj.ids()
        npt.assert_equal(tested, np.array([0, 1, 2, 3]))
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        assert self.test_obj.ids(0) == [0]
        assert self.test_obj.ids(np.int64(0)) == [0]
        assert self.test_obj.ids(np.uint64(0)) == [0]
        assert self.test_obj.ids(np.int32(0)) == [0]
        assert self.test_obj.ids(np.uint32(0)) == [0]
        npt.assert_equal(self.test_obj.ids([0, 1]), np.array([0, 1]))
        npt.assert_equal(self.test_obj.ids(np.array([0, 1])), np.array([0, 1]))
        npt.assert_equal(self.test_obj.ids(CircuitEdgeId("default", 0)), [0])
        npt.assert_equal(self.test_obj.ids(CircuitEdgeId("default2", 0)), [])
        ids = CircuitEdgeIds.from_tuples([("default", 0), ("default", 1)])
        npt.assert_equal(self.test_obj.ids(ids), np.array([0, 1]))
        ids = CircuitEdgeIds.from_tuples([("default", 0), ("default2", 1)])
        npt.assert_equal(self.test_obj.ids(ids), np.array([0]))
        ids = CircuitEdgeIds.from_tuples([("default2", 0), ("default2", 1)])
        npt.assert_equal(self.test_obj.ids(ids), [])
        npt.assert_equal(self.test_obj.ids(), np.array([0, 1, 2, 3]))

        # limit too big compared to the number of ids
        npt.assert_equal(self.test_obj.ids(limit=15), [0, 1, 2, 3])
        npt.assert_equal(len(self.test_obj.ids(sample=2)), 2)
        # if sample > population.size --> sample = population.size
        npt.assert_equal(len(self.test_obj.ids(sample=25)), 4)

    def test_get_1(self):
        properties = [
            Synapse.PRE_GID,
            Synapse.POST_GID,
            Synapse.AXONAL_DELAY,
            Synapse.POST_X_CENTER,
            test_module.DYNAMICS_PREFIX + "param1",
        ]
        edge_ids = [0, 1]
        actual = self.test_obj.get(edge_ids, properties)
        expected = pd.DataFrame(
            [
                (2, 0, 99.8945, 1110.0, 0.0),
                (0, 1, 88.1862, 1111.0, 1.0),
            ],
            columns=properties,
            index=index_as_ids_dtypes(edge_ids),
        )
        pdt.assert_frame_equal(actual, expected, check_dtype=False)

    def test_get_2(self):
        prop = Synapse.AXONAL_DELAY
        edge_ids = [1, 0]
        actual = self.test_obj.get(edge_ids, prop)
        expected = pd.Series([88.1862, 99.8945], index=index_as_ids_dtypes(edge_ids), name=prop)
        pdt.assert_series_equal(actual, expected, check_dtype=False)

    def test_get_3(self):
        properties = [Synapse.PRE_GID, Synapse.AXONAL_DELAY]
        pdt.assert_series_equal(
            self.test_obj.get([], properties[0]), pd.Series(name=properties[0], dtype=np.float64)
        )
        pdt.assert_frame_equal(self.test_obj.get([], properties), pd.DataFrame(columns=properties))

    def test_get_4(self):
        with pytest.raises(BluepySnapError):
            self.test_obj.get([0], "no-such-property")

    def test_get_without_properties(self):
        edge_ids = [0, 1]
        actual = self.test_obj.get(edge_ids, None)
        expected = np.asarray(edge_ids, dtype=np.int64)
        npt.assert_equal(actual, expected)

    def test_get_all_edge_ids_types(self):
        assert self.test_obj.get(0, Synapse.PRE_GID).tolist() == [2]
        assert self.test_obj.get(np.int64(0), Synapse.PRE_GID).tolist() == [2]
        assert self.test_obj.get(np.uint64(0), Synapse.PRE_GID).tolist() == [2]
        assert self.test_obj.get(np.int32(0), Synapse.PRE_GID).tolist() == [2]
        assert self.test_obj.get(np.uint32(0), Synapse.PRE_GID).tolist() == [2]

        assert self.test_obj.get([0], Synapse.PRE_GID).tolist() == [2]
        assert self.test_obj.get([0, 1], Synapse.PRE_GID).tolist() == [2, 0]
        assert self.test_obj.get(CircuitEdgeId("default", 0), Synapse.PRE_GID).tolist() == [2]
        assert self.test_obj.get(
            CircuitEdgeIds.from_tuples([("default", 0)]), Synapse.PRE_GID
        ).tolist() == [2]
        assert self.test_obj.get(CircuitEdgeId("default2", 0), Synapse.PRE_GID).tolist() == []
        assert self.test_obj.get(
            CircuitEdgeIds.from_tuples([("default", 0), ("default", 1)]), Synapse.PRE_GID
        ).tolist() == [2, 0]
        assert self.test_obj.get(
            CircuitEdgeIds.from_tuples([("default", 0), ("default2", 1)]), Synapse.PRE_GID
        ).tolist() == [2]
        assert (
            self.test_obj.get(
                CircuitEdgeIds.from_tuples([("default2", 0), ("default2", 1)]), Synapse.PRE_GID
            ).tolist()
            == []
        )

    def test_positions_1(self):
        actual = self.test_obj.positions([0], "afferent", "center")
        expected = pd.DataFrame(
            [[1110.0, 1120.0, 1130.0]], index=index_as_ids_dtypes([0]), columns=["x", "y", "z"]
        )
        pdt.assert_frame_equal(actual, expected)
        pdt.assert_frame_equal(self.test_obj.positions(0, "afferent", "center"), actual)
        pdt.assert_frame_equal(self.test_obj.positions(np.int64(0), "afferent", "center"), actual)
        pdt.assert_frame_equal(self.test_obj.positions(np.uint64(0), "afferent", "center"), actual)
        pdt.assert_frame_equal(self.test_obj.positions(np.int32(0), "afferent", "center"), actual)
        pdt.assert_frame_equal(self.test_obj.positions(np.uint32(0), "afferent", "center"), actual)

    def test_positions_2(self):
        actual = self.test_obj.positions([1], "afferent", "surface")
        expected = pd.DataFrame(
            [[1211.0, 1221.0, 1231.0]], index=index_as_ids_dtypes([1]), columns=["x", "y", "z"]
        )
        pdt.assert_frame_equal(actual, expected)

    def test_positions_3(self):
        actual = self.test_obj.positions([2], "efferent", "center")
        expected = pd.DataFrame(
            [[2112.0, 2122.0, 2132.0]], index=index_as_ids_dtypes([2]), columns=["x", "y", "z"]
        )
        pdt.assert_frame_equal(actual, expected)

    def test_positions_4(self):
        actual = self.test_obj.positions([3], "efferent", "surface")
        expected = pd.DataFrame(
            [[2213.0, 2223.0, 2233.0]], index=index_as_ids_dtypes([3]), columns=["x", "y", "z"]
        )
        pdt.assert_frame_equal(actual, expected)

    def test_positions_5(self):
        with pytest.raises(AssertionError):
            self.test_obj.positions([2], "err", "center")

    def test_positions_6(self):
        with pytest.raises(AssertionError):
            self.test_obj.positions([2], "afferent", "err")

    def test_afferent_nodes(self):
        tested = self.test_obj.afferent_nodes(0)
        npt.assert_equal(tested, [2])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        tested = self.test_obj.afferent_nodes(np.int64(0))
        npt.assert_equal(tested, [2])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        tested = self.test_obj.afferent_nodes(np.uint64(0))
        npt.assert_equal(tested, [2])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        tested = self.test_obj.afferent_nodes(np.int32(0))
        npt.assert_equal(tested, [2])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        tested = self.test_obj.afferent_nodes(np.uint32(0))
        npt.assert_equal(tested, [2])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        npt.assert_equal(self.test_obj.afferent_nodes(0, unique=False), [2])

        npt.assert_equal(self.test_obj.afferent_nodes(1), [0, 2])
        npt.assert_equal(self.test_obj.afferent_nodes(1, unique=False), [0, 0, 2])

        npt.assert_equal(self.test_obj.afferent_nodes(2), [])

        npt.assert_equal(self.test_obj.afferent_nodes([0, 1]), [0, 2])
        npt.assert_equal(self.test_obj.afferent_nodes([0, 1], unique=False), [2, 0, 0, 2])

        npt.assert_equal(self.test_obj.afferent_nodes({}), [0, 2])
        npt.assert_equal(
            self.test_obj.afferent_nodes({"mtype": "L2_X"}), [2]
        )  # eq node id 0 as target
        npt.assert_equal(
            self.test_obj.afferent_nodes({"mtype": "L2_X"}), [2]
        )  # eq node id 0 as target

        npt.assert_equal(self.test_obj.afferent_nodes(None), [0, 2])
        npt.assert_equal(self.test_obj.afferent_nodes(None, unique=False), [2, 0, 0, 2])

    def test_efferent_nodes(self):
        tested = self.test_obj.efferent_nodes(0)
        npt.assert_equal(tested, [1])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        tested = self.test_obj.efferent_nodes(np.int64(0))
        npt.assert_equal(tested, [1])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        tested = self.test_obj.efferent_nodes(np.uint64(0))
        npt.assert_equal(tested, [1])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        tested = self.test_obj.efferent_nodes(np.int32(0))
        npt.assert_equal(tested, [1])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        tested = self.test_obj.efferent_nodes(np.uint32(0))
        npt.assert_equal(tested, [1])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        npt.assert_equal(self.test_obj.efferent_nodes(0, unique=False), [1, 1])

        npt.assert_equal(self.test_obj.efferent_nodes(1), [])
        npt.assert_equal(self.test_obj.efferent_nodes(1, unique=False), [])

        npt.assert_equal(self.test_obj.efferent_nodes(2), [0, 1])
        npt.assert_equal(self.test_obj.efferent_nodes(2, unique=False), [0, 1])

        npt.assert_equal(self.test_obj.efferent_nodes([0, 1]), [1])
        npt.assert_equal(self.test_obj.efferent_nodes([0, 1], unique=False), [1, 1])

        npt.assert_equal(self.test_obj.efferent_nodes({}), [0, 1])
        npt.assert_equal(
            self.test_obj.efferent_nodes({"mtype": "L2_X"}), [1]
        )  # eq node id 0 as source
        npt.assert_equal(
            self.test_obj.efferent_nodes({"mtype": "L2_X"}), [1]
        )  # eq node id 0 as source

        npt.assert_equal(self.test_obj.efferent_nodes(None), [0, 1])
        npt.assert_equal(self.test_obj.efferent_nodes(None, unique=False), [0, 1, 1, 1])

    def test_afferent_edges(self):
        tested = self.test_obj.afferent_edges([0, 1], None)
        npt.assert_equal(tested, [0, 1, 2, 3])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

    def test_afferent_edges_1(self):
        npt.assert_equal(self.test_obj.afferent_edges(1, None), [1, 2, 3])

        npt.assert_equal(self.test_obj.afferent_edges(np.int64(1), None), [1, 2, 3])

        npt.assert_equal(self.test_obj.afferent_edges(np.uint64(1), None), [1, 2, 3])

        npt.assert_equal(self.test_obj.afferent_edges(np.int32(1), None), [1, 2, 3])

        npt.assert_equal(self.test_obj.afferent_edges(np.uint32(1), None), [1, 2, 3])

    def test_afferent_edges_2(self):
        properties = [Synapse.AXONAL_DELAY]
        pdt.assert_frame_equal(
            self.test_obj.afferent_edges(1, properties),
            pd.DataFrame(
                [
                    [88.1862],
                    [52.1881],
                    [11.1058],
                ],
                columns=properties,
                index=index_as_ids_dtypes([1, 2, 3]),
            ),
            check_dtype=False,
        )

    def test_efferent_edges_1(self):
        tested = self.test_obj.efferent_edges(2, None)
        npt.assert_equal(tested, [0, 3])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

    def test_efferent_edges_2(self):
        properties = [Synapse.AXONAL_DELAY]
        pdt.assert_frame_equal(
            self.test_obj.efferent_edges(2, properties),
            pd.DataFrame(
                [
                    [99.8945],
                    [11.1058],
                ],
                columns=properties,
                index=index_as_ids_dtypes([0, 3]),
            ),
            check_dtype=False,
        )

    def test_pair_edges_1(self):
        tested = self.test_obj.pair_edges(0, 2, None)
        npt.assert_equal(tested, [])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

    def test_pair_edges_2(self):
        actual = self.test_obj.pair_edges(0, 2, [Synapse.AXONAL_DELAY])
        assert actual.empty

    def test_pair_edges_3(self):
        tested = self.test_obj.pair_edges(2, 0, None)
        npt.assert_equal(tested, [0])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

    def test_pair_edges_4(self):
        properties = [Synapse.AXONAL_DELAY]
        pdt.assert_frame_equal(
            self.test_obj.pair_edges(2, 0, properties),
            pd.DataFrame(
                [
                    [99.8945],
                ],
                columns=properties,
                index=index_as_ids_dtypes([0]),
            ),
            check_dtype=False,
        )

    def test_pathway_edges_1(self):
        properties = [Synapse.AXONAL_DELAY]
        pdt.assert_frame_equal(
            self.test_obj.pathway_edges([0, 1], [1, 2], properties),
            pd.DataFrame(
                [
                    [88.1862],
                    [52.1881],
                ],
                columns=properties,
                index=index_as_ids_dtypes([1, 2]),
            ),
            check_dtype=False,
        )

    def test_pathway_edges_2(self):
        tested = self.test_obj.pathway_edges([1, 2], [0, 2], None)
        npt.assert_equal(tested, [0])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

    def test_pathway_edges_3(self):
        npt.assert_equal(self.test_obj.pathway_edges([0, 1], None, None), [1, 2])

    def test_pathway_edges_4(self):
        npt.assert_equal(self.test_obj.pathway_edges(None, [0, 1], None), [0, 1, 2, 3])

    def test_pathway_edges_5(self):
        with pytest.raises(BluepySnapError):
            self.test_obj.pathway_edges(None, None, None)

    def test_pathway_edges_6(self):
        ids = CircuitNodeIds.from_dict({"default": [0, 1]})
        npt.assert_equal(self.test_obj.pathway_edges(ids, None, None), [1, 2])

    def test_iter_connections_1(self):
        it = self.test_obj.iter_connections([0, 2], [1])
        assert next(it) == (0, 1)
        assert next(it) == (2, 1)
        with pytest.raises(StopIteration):
            next(it)

    def test_iter_connections_2(self):
        it = self.test_obj.iter_connections([0, 2], [1], unique_node_ids=True)
        assert list(it) == [(0, 1)]

    def test_iter_connections_3(self):
        it = self.test_obj.iter_connections([0, 2], [1], shuffle=True)
        assert sorted(it) == [(0, 1), (2, 1)]

    def test_iter_connections_4(self):
        it = self.test_obj.iter_connections(None, None)
        with pytest.raises(BluepySnapError):
            next(it)

    def test_iter_connections_5(self):
        it = self.test_obj.iter_connections(None, [1])
        assert list(it) == [(0, 1), (2, 1)]

    def test_iter_connections_6(self):
        it = self.test_obj.iter_connections([2], None)
        assert list(it) == [(2, 0), (2, 1)]

    def test_iter_connections_7(self):
        it = self.test_obj.iter_connections([], [0, 1, 2])
        assert list(it) == []

    def test_iter_connections_8(self):
        it = self.test_obj.iter_connections([0, 2], [1], return_edge_ids=True)
        npt.assert_equal(list(it), [(0, 1, [1, 2]), (2, 1, [3])])

    def test_iter_connections_9(self):
        it = self.test_obj.iter_connections([0, 2], [1], return_edge_count=True)
        assert list(it) == [(0, 1, 2), (2, 1, 1)]

    def test_iter_connections_10(self):
        with pytest.raises(BluepySnapError):
            self.test_obj.iter_connections(
                [0, 2], [1], return_edge_ids=True, return_edge_count=True
            )

    def test_iter_connection_unique(self):
        with copy_test_data() as (config_dir, config_path):
            with edit_config(config_path) as config:
                config["networks"]["edges"] = [
                    {
                        "edge_types_file": None,
                        "edges_file": str(Path(config_dir) / "edges_complete_graph.h5"),
                        "populations": {"default": {"type": "chemical"}},
                    }
                ]

            test_obj = TestEdgePopulation.get_edge_population(config_path, "default")

            it = test_obj.iter_connections([0, 1, 2], [0, 1, 2])
            assert sorted(it) == [(0, 1), (0, 2), (1, 0), (1, 2), (2, 0), (2, 1)]

            it = test_obj.iter_connections([0, 1, 2], [0, 1, 2], unique_node_ids=True)
            assert sorted(it) == [(0, 1), (1, 0)]

            it = test_obj.iter_connections([0, 1, 2], [0, 2], unique_node_ids=True)
            assert sorted(it) == [(0, 2), (1, 0)]

            it = test_obj.iter_connections([0, 2], [0, 2], unique_node_ids=True)
            assert sorted(it) == [(0, 2), (2, 0)]

            it = test_obj.iter_connections([0, 1, 2], [0, 2, 1], unique_node_ids=True)
            assert sorted(it) == [(0, 1), (1, 0)]

            it = test_obj.iter_connections([1, 2], [0, 1, 2], unique_node_ids=True)
            assert sorted(it) == [(1, 0), (2, 1)]

            it = test_obj.iter_connections([0, 1, 2], [1, 2], unique_node_ids=True)
            assert sorted(it) == [(0, 1), (1, 2)]

    def test_h5_filepath_from_config(self):
        assert self.test_obj.h5_filepath == str(TEST_DATA_DIR / "edges.h5")

    @pytest.mark.skip(reason="Until spatial-index is released publicly")
    def test_spatial_synapse_index(self):
        with mock.patch("spatial_index.open_index") as mock_open_index:
            self.test_obj.spatial_synapse_index
        mock_open_index.assert_called_once_with("path/to/edge/dir")

    @mock.patch.dict(sys.modules, {"spatial_index": mock.Mock()})
    def test_spatial_synapse_index_call(self):
        with pytest.raises(
            BluepySnapError,
            match="It appears default does not have synapse indices",
        ):
            self.test_obj.spatial_synapse_index

    def test_spatial_synapse_index_error(self):
        with pytest.raises(
            BluepySnapError,
            match=(
                "Spatial index is for now only available internally to BBP. "
                "It requires `spatial_index`, an internal package."
            ),
        ):
            self.test_obj.spatial_synapse_index

    def test_pickle(self, tmp_path):
        pickle_path = tmp_path / "pickle.pkl"

        # trigger some cached properties, to makes sure they aren't being pickeld
        self.test_obj.source
        self.test_obj.target
        self.test_obj.property_dtypes

        with open(pickle_path, "wb") as fd:
            pickle.dump(self.test_obj, fd)

        with open(pickle_path, "rb") as fd:
            edge_population = pickle.load(fd)

        assert pickle_path.stat().st_size < 130 + PICKLED_SIZE_ADJUSTMENT
        assert edge_population.name == "default"


class TestEdgePopulationSpatialIndex:
    def setup_method(self):
        self.test_obj = TestEdgePopulation.get_edge_population(
            TEST_DATA_DIR / "circuit_config.json", "default2"
        )

    @mock.patch.dict(sys.modules, {"spatial_index": mock.Mock()})
    def test_spatial_synapse_index_call(self):
        self.test_obj.spatial_synapse_index
        mock = sys.modules["spatial_index"].open_index
        assert mock.call_count == 1
        assert mock.call_args[0][0].endswith("path/to/edge/dir")



# Copyright (c) 2019, EPFL/Blue Brain Project
#
# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Simulation and Neural network Analysis Productivity layer."""

from bluepysnap.circuit import Circuit
from bluepysnap.config import Config
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.simulation import Simulation

try:
    from bluepysnap._version import version as __version__
except ImportError:  # pragma: no cover
    __version__ = "unknown"
